import time

from Foundation import NSAppleScript
from Foundation import NSAppleScriptErrorBriefMessage
from Foundation import NSAppleScriptErrorMessage
from Foundation import NSAppleScriptErrorNumber

from macuitest.lib.applescript_lib.aeconverter import ae_converter


class AppleScriptError(Exception):
    """Indicates an AppleScript compilation/execution error."""

    def __init__(self, error_info):
        self._error_info = dict(error_info)

    def __repr__(self):
        return f"{self.__class__.__name__}({self._error_info})"

    @property
    def message(self) -> str:
        msg = self._error_info.get(NSAppleScriptErrorMessage)
        if not msg:
            msg = self._error_info.get(NSAppleScriptErrorBriefMessage, "Script Error")
        return msg

    @property
    def number(self):
        """ "int | None -- the error number, if given")"""
        return self._error_info.get(NSAppleScriptErrorNumber)


class AppleScriptWrapper:
    """Wrapper for AppleScript with a set of easy to use methods."""

    allowed_modifier_keys = ("control", "shift", "option", "command")
    key_codes: dict = {
        "return": 36,
        "esc": 53,
        "delete": 51,
        "left": 123,
        "right": 124,
        "up": 126,
        "down": 125,
        "tab": 48,
    }
    __pass_as_key_code = {'"': (39, True), "'": (39, False)}

    def typewrite(self, phrase: str) -> None:
        """Type `phrase` with a delay between key presses."""
        for char in phrase:
            time.sleep(0.005)
            if char in self.__pass_as_key_code:
                key_code, is_shift_required = self.__pass_as_key_code[char]
                self.send_keycode(key_code, "shift") if is_shift_required else self.send_keycode(
                    key_code
                )
            else:
                self.send_keystroke(char)

    def send_keystroke(self, phrase: str, *args) -> None:
        """Send keystroke event with the specified `phrase` (type it using AppleScript)."""
        self.__send_event("keystroke", phrase, *args)

    def send_keycode(self, key_code: int, *args) -> None:
        """Send keystroke event with the specified `key_code` (type it using AppleScript)."""
        self.__send_event("key code", key_code, *args)

    def __send_event(self, event_type: str, message: [str, int], *args):
        for modifier in args:
            if modifier not in self.allowed_modifier_keys:
                raise KeyError(f'{modifier} is not a modifier key.')
        _cmd = f'{event_type} "{message}"'
        if args:
            _cmd = (
                f'{event_type} "{message}" using '
                f'{{{", ".join([f"{modifier_key} down" for modifier_key in args])}}}'
            )
        return self.tell_sys_events(_cmd)

    def tell_app_process(self, command: str, app_process: str):
        return self.execute(
            f'tell app "System Events" to tell application process "{app_process}" to {command}'
        )

    def tell_app(self, app: str, command: str, ignoring_responses: bool = False):
        _tell_what = (
            f'tell application "{app}" to {command}'
            if not ignoring_responses
            else f'ignoring application responses\ntell application "{app}" '
            f"to {command}\nend ignoring"
        )
        return self.execute(_tell_what)

    def tell_sys_events(self, command: str):
        return self.execute(f'tell application "System Events" to {command}')

    @staticmethod
    def execute(cmd: str):
        """Execute AppleScript command abd returns exitcode, stdout and stderr.
        :param str cmd: apple script
        :return: exitcode, stdout and stderr"""
        result, error = NSAppleScript.alloc().initWithSource_(cmd).executeAndReturnError_(None)
        if error:
            raise AppleScriptError(error)
        return ae_converter.unpack(result)


as_wrapper = AppleScriptWrapper()



import time
from dataclasses import dataclass

from macuitest.lib.elements.controllers.keyboard_controller import keyboard
from macuitest.lib.elements.controllers.mouse_controller import MouseController


@dataclass(frozen=True)
class MouseConfig:
    """Mouse options."""

    move: float = 0.18  # Cursor roaming time.
    hold: float = 0.24  # Time to hold a button selected.
    pause: float = 0.24  # Pause after an action.
    default_position: tuple = (5, 3)


class Mouse:
    """Cursor manipulator."""

    def __init__(self, controller: MouseController):
        self.controller = controller

    def paste(self, x: int, y: int, phrase: str = "") -> None:
        """Hover over the position and click once. Then paste `phrase` from clipboard."""
        self.click(x, y)
        keyboard.write(phrase, pause=0.02)

    def double_click(
        self, x: int, y: int, _x: int = 0, _y: int = 0, duration: float = MouseConfig.move
    ) -> None:
        """Hover over position and click twice."""
        self.hover(x, y, duration=duration)
        self.controller.multi_click(x, y, button="left", clicks=2)

    def drag(self, x1: int, y1: int, x2: int, y2: int) -> None:
        self.hover(x1, y1)
        self.controller.mouse_down(x1, y1, "left")
        time.sleep(0.5)
        self.controller.drag_to(x2, y2)
        time.sleep(0.5)
        self.controller.mouse_up(x2, y2, "left")
        time.sleep(0.25)

    def click(
        self,
        x: int,
        y: int,
        hold: float = MouseConfig.hold,
        duration: float = MouseConfig.move,
        pause: float = MouseConfig.pause,
    ) -> None:
        """Hover over position and left-click once."""
        time.sleep(pause)
        self.hover(x, y, duration)
        self._press_mouse_button(x, y, mouse_button="left", hold=hold, pause=0.125)

    def right_click(
        self,
        x: int,
        y: int,
        hold: float = MouseConfig.hold,
        duration: float = MouseConfig.move,
        pause: float = MouseConfig.pause,
    ) -> None:
        """Hover over the position and control-click once."""
        time.sleep(pause)
        self.hover(x, y, duration)
        self._press_mouse_button(x, y, mouse_button="right", hold=hold, pause=0.125)

    def scroll(self, x: int, y: int, scrolls: int = 1) -> None:
        self.hover(x, y)
        self.controller.vertical_scroll(scrolls)

    def reset(self):
        self.hover(*MouseConfig.default_position)

    def hover(self, x: int, y: int, duration: float = MouseConfig.move) -> None:
        """Hover over the position."""
        self.controller.move_to(x, y, duration=duration)

    def _press_mouse_button(self, x: int, y: int, mouse_button: str, hold: float, pause: float):
        time.sleep(pause)
        self.controller.mouse_down(x, y, mouse_button)
        time.sleep(hold)
        self.controller.mouse_up(x, y, mouse_button)
        time.sleep(0.25)  # We want to wait a bit for system to register the event.


mouse = Mouse(MouseController())



import datetime
import hashlib
import math
import os
import shutil
import zipfile
from pathlib import Path
from pwd import getpwuid
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union


class FileManager:
    permissions = {
        "all": 0o777,
        "default": 0o755,
        "read_only": 0o444,
        "write_only": 0o222,
        "no_access": 0o000,
    }

    def __init__(self, executor):
        self.executor = executor

    def clear_quarantine_attribute(self, path: str):
        self.executor.sudo(f'xattr -r -d com.apple.quarantine "{path}"')

    def read_metadata(self, path: str, key: Optional[str] = None):
        if key is None:
            return self.executor.get_output(f'mdls "{path}"')
        return self.executor.get_output(f'mdls -raw -name {key} "{path}"')

    def touch(self, path):
        """
        Sets the modification and access times of files.
        If any file does not exist, it is created with default permissions.
        """
        self.executor.execute(f'touch "{path}"')

    def dump_text_to_file(self, text: str, file_: str):
        command = f'printf "{text}" > "{file_}"'
        if self.executor.execute(command).returncode != 0:
            self.executor.sudo(command)

    def make_file(self, path: str, size: str = "8m", sudo: bool = False):
        """Create a file with size `size`[b|k|m|g] in `path`."""
        execute = self.executor.sudo if sudo else self.executor.execute
        return execute(f'mkfile {size} "{path}"')

    def make_dir(self, path: str, sudo: bool = False):
        execute = self.executor.sudo if sudo else self.executor.execute
        return execute(f'mkdir -p "{path}"')

    def last_modified__timedelta(self, path: str):
        timedelta = datetime.datetime.now() - datetime.datetime.fromtimestamp(
            self.last_modified(path)
        )
        return int(timedelta.total_seconds())

    @staticmethod
    def last_modified(path: str):
        return os.path.getmtime(path)

    @staticmethod
    def is_path_exists(path: str) -> bool:
        return os.path.exists(path)

    @staticmethod
    def is_dir(path: str) -> bool:
        return os.path.isdir(path)

    @staticmethod
    def is_file(path: str) -> bool:
        return os.path.isfile(path)

    @staticmethod
    def read_file(source: str) -> str:
        with open(source, "r", encoding="ISO-8859-1") as f:
            return f.read()

    @staticmethod
    def write_to_file(source: str, data: str) -> None:
        with open(source, "w") as f:
            f.write(data)

    def move_tree(self, source, destination):
        """Move `source` file to `destination`."""
        self.copy_tree(source, destination)
        self.remove_tree(source)

    def remove_content_from(self, content, directory) -> None:
        """Remove `content` from `directory`"""
        if not content:
            return
        for item in content:
            self.remove_tree(os.path.join(directory, item))

    @staticmethod
    def rename_file(default_file_path, updated_file_path):
        return os.rename(default_file_path, updated_file_path)

    def clear_directory(
        self, directory: str, pattern: str = "*", sudo: bool = False
    ):  # Clear all files, by default.
        return (self.executor.sudo if sudo else self.executor.execute)(
            f"rm -rf {directory}/{pattern}"
        )

    def remove_tree(self, tree_path):
        """Delete a directory tree or a file."""
        if not os.path.exists(tree_path):
            return
        try:
            (shutil.rmtree if os.path.isdir(tree_path) else os.unlink)(tree_path)
        except (OSError, PermissionError):
            self.executor.sudo(f'rm -r "{tree_path}"')

    def copy_file(self, source: str, destination: str):
        self.executor.execute(f'cp "{source}" "{destination}"')

    def move_file(self, source: str, destination: str):
        self.executor.execute(f'mv "{source}" "{destination}"')

    def copy_tree(self, source: str, destination: str):
        """Copy a directory tree or folder's content to the destination."""
        self.makedir(destination)
        execute = (
            self.executor.sudo
            if self.__are_elevated_rights_needed(destination)
            else self.executor.execute
        )
        execute(f'rsync -q -ahP "{source}" "{destination}"')

    def __are_elevated_rights_needed(self, destination: Union[Path, str]):
        result = False
        protected_folders = ("dev", "var", "usr", "private", "Library", "System")
        _destination_ = str(destination)
        if (
            self.get_file_owner(_destination_) == "root"
            or _destination_.split("/")[1] in protected_folders
        ):
            result = True
        return result

    def get_size(self, file_path) -> int:
        """Total size, in KB"""
        if Path(file_path).exists():
            return int(self.executor.get_output(f'du -hk "{file_path}"').split()[0].strip())
        raise FileNotFoundError(file_path)

    def folder_newest_file(self, folder: str) -> str:
        if Path(folder).is_dir():
            return max(self.folder_content(folder), key=os.path.getctime)
        raise NotADirectoryError(folder)

    @staticmethod
    def folder_content(folder: str) -> List[str]:
        if Path(folder).is_dir():
            return [
                path_.as_posix()
                for path_ in Path(folder).glob("*")
                if not path_.name.startswith(".")
            ]
        raise NotADirectoryError(folder)

    def get_bundle_content_owners(self, path: str) -> List[Tuple[str, str, str]]:
        raw = (line for line in self.executor.get_output(f'ls -R -l "{path}"').splitlines())
        payload = (
            line.split()
            for line in raw
            if line and not (line.startswith("total") or line.startswith("/"))
        )
        return [(line[-1], line[2], line[3]) for line in payload]

    @staticmethod
    def get_file_owner(file_path):
        try:
            return getpwuid(os.stat(file_path).st_uid).pw_name
        except OSError as e:
            if e.errno == 2:
                return "missing"
            else:
                raise

    @staticmethod
    def unzip_file(src, dst):
        with zipfile.ZipFile(src) as z:
            z.extractall(dst)

    def unzip_app(self, src, dst):
        if not os.path.exists(dst):
            os.makedirs(dst, exist_ok=True)
        self.executor.execute(f'unzip -qq -o "{src}" -d "{dst}"')

    @staticmethod
    def zip_file(src, dst):
        relroot = os.path.abspath(os.path.join(src, os.pardir))
        with zipfile.ZipFile(dst, "w", zipfile.ZIP_DEFLATED) as zip_:
            for root, dirs, files in os.walk(src):
                zip_.write(root, os.path.relpath(root, relroot))
                for file_ in files:
                    filename = os.path.join(root, file_)
                    if os.path.isfile(filename):
                        arcname = os.path.join(os.path.relpath(root, relroot), file_)
                        zip_.write(filename, arcname)

    def makedir(self, destination):
        if not os.path.exists(destination):
            if self.executor.execute(f'mkdir -p "{destination}"').returncode != 0:
                self.executor.sudo(f'mkdir -p "{destination}"')

    def lock_file(self, path: str):
        if isinstance(path, Path):
            path = path.as_posix()
        return self.executor.sudo(f'chflags uchg "{path}"')

    def change_file_owner(self, file: str, owner: str):
        return self.executor.sudo(f'chown {owner} "{file}"')

    def change_file_permissions(self, path: str, permissions, sudo: bool = False):
        """Change file permissions."""
        (self.executor.sudo if sudo else self.executor.execute)(f'chmod {permissions} "{path}"')

    @staticmethod
    def change_files_permissions(folder, permissions):
        """Change file permissions in `folder`."""
        for root, subfolders, files in os.walk(folder):
            [os.chmod(os.path.join(root, file_), permissions) for file_ in files]

    @staticmethod
    def convert_file_size_to_human_readable(size):
        if size == "":
            result = 0
        else:
            size = float(size)
            if size == 0:
                return "0B"
            size_name = ("bytes", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
            i = int(math.floor(math.log(size, 1000)))
            p = math.pow(1000, i)
            s = round(size / p, 2)
            result = "{} {}".format(s, size_name[i])
        return result

    @staticmethod
    def convert_file_size_to_bytes(string):
        _size, _unit = string.strip().lower().replace("zero", "0").replace(",", "").split()
        return int(float(_size) * 1000 ** ("bytes", "kb", "mb", "gb", "tb").index(_unit))

    def find_duplicates_number(self, folders):
        """Return duplicates in format {filename:number of duplicates}"""
        if isinstance(folders, str):
            folders = [
                folders,
            ]
        hashes: dict = dict()
        duplicates: dict = dict()
        for folder in folders:
            for dirpath, dirnames, files_list in os.walk(folder):
                for filename in files_list:
                    if not filename.startswith("."):
                        full_path = os.path.join(dirpath, filename)
                        if not os.path.islink(full_path) and os.path.exists(full_path):
                            file_id = self._generate_md5_checksum(full_path)
                            duplicate = hashes.get(file_id)
                            if duplicate:
                                counter = duplicates.get(filename, 1)
                                duplicates[filename] = counter + 1
                                hashes[file_id].append(full_path)
                            else:
                                hashes[file_id] = [full_path]
        return duplicates

    @staticmethod
    def _generate_md5_checksum(file_name, chunk_size=1024):
        """Calculate md5 checksum of `file_name`."""
        hash_ = hashlib.md5()
        with open(file_name, "rb") as f:
            chunk = f.read(chunk_size)
            while chunk:
                hash_.update(chunk)
                chunk = f.read(chunk_size)
        return hash_.hexdigest()



from typing import ClassVar
from typing import Dict

from macuitest.lib.operating_system.db_manager import DataBaseManager
from macuitest.lib.operating_system.env import env


class TCCManager:
    """
    TCCManager manages the privacy database, which stores decisions
    the user has made about whether apps may access personal data.
    """

    sys_db: ClassVar[str] = f"{env.system_app_support}/com.apple.TCC/TCC.db"
    usr_db: ClassVar[str] = f"{env.home}{sys_db}"
    services: ClassVar[Dict[str, str]] = {
        "full_disk_access": "kTCCServiceSystemPolicyAllFiles",
        "camera_access": "kTCCServiceCamera",
        "documents_folder_access": "kTCCServiceSystemPolicyDocumentsFolder",
        "desktop_folder_access": "kTCCServiceSystemPolicyDesktopFolder",
        "downloads_folder_access": "kTCCServiceSystemPolicyDownloadsFolder",
    }

    def __init__(self, domain: str = "system"):
        self.db: str = self.sys_db if domain == "system" else self.usr_db
        self.db_manager = DataBaseManager(self.db)

    def list_access_policies(self, client: str) -> Dict[str, int]:
        is_allowed = "allowed" if env.version < (11, 0) else "auth_value"
        query = f'SELECT service, {is_allowed} FROM access WHERE client is "{client}"'
        return dict(self.db_manager.cursor.execute(query).fetchall())

    def revoke_access_to_service(self, service: str, client: str) -> None:
        query = f'DELETE FROM access WHERE service is "{service}" AND client is "{client}"'
        self.db_manager.cursor.execute(query)
        self.db_manager.connection.commit()



from dataclasses import dataclass
from typing import List

import AppKit
from ApplicationServices import AXUIElementCreateApplication
from ApplicationServices import AXUIElementCreateSystemWide

from macuitest.lib.elements.native.calls import AXError
from macuitest.lib.elements.native.calls import AXErrorAttributeUnsupported
from macuitest.lib.elements.native.calls import AXErrorNoValue
from macuitest.lib.elements.native.calls import AXErrorUnsupported
from macuitest.lib.elements.native.calls import check_attribute_settable
from macuitest.lib.elements.native.calls import get_accessibility_element_attribute
from macuitest.lib.elements.native.calls import get_accessibility_object_on_screen_position
from macuitest.lib.elements.native.calls import get_accessibility_object_pid
from macuitest.lib.elements.native.calls import get_element_action_names
from macuitest.lib.elements.native.calls import get_element_attribute_names
from macuitest.lib.elements.native.calls import get_running_apps
from macuitest.lib.elements.native.calls import get_running_apps_with_bundle_id
from macuitest.lib.elements.native.calls import perform_action_on_element
from macuitest.lib.elements.native.calls import set_accessibility_api_timeout
from macuitest.lib.elements.native.calls import set_attribute_value
from macuitest.lib.elements.native.converter import Converter


@dataclass
class NSApplicationActivationOptions:
    Default: int = 1
    ActivateAllWindows: int = 2
    ActivateIgnoringOtherWindows: int = 3


class NativeUIElement:
    def __init__(self, ref=None):
        self.ref = ref
        self.converter = Converter(self.__class__)
        self.__application = None

    def __repr__(self):
        for identifier in ("AXTitle", "AXValue", "AXRoleDescription"):
            title = self.get_ax_attribute(identifier)
            if title:
                title = f'"{title}"'
                break
        return f'NativeUIElement {self.get_ax_attribute("AXRole")} {title}'.strip()

    def __dir__(self):
        return self.ax_attributes + self.ax_actions + list(self.__dict__.keys()) + dir(super())

    @classmethod
    def from_bundle_id(cls, bundle_id: str):
        """Get application by the specified bundle ID."""
        matches = get_running_apps_with_bundle_id(bundle_id)
        if not matches:
            raise ValueError(f'"{bundle_id}" not found among running apps.')
        return cls.from_pid(matches[0].processIdentifier())

    @classmethod
    def from_localized_name(cls, name: str):
        """Get the application by the specified localized name. Wildcards are also allowed."""
        for app in get_running_apps():
            if app.localizedName() == name:
                return cls.from_pid(app.processIdentifier())
        raise ValueError(f'"{name}" not found among running applications.')

    @classmethod
    def from_pid(cls, pid: int):
        """Get application by the specified process ID."""
        return cls(ref=AXUIElementCreateApplication(pid))

    @classmethod
    def systemwide(cls):
        """Create an instance with the AXUIElementRef for the system-wide accessibility object."""
        return cls(ref=AXUIElementCreateSystemWide())

    def get_menu_item(self, menu: str, menu_item: str):
        """Return the specified menu item."""
        for item in self.get_menu_items(menu):
            if item.get_ax_attribute("AXTitle") == menu_item:
                return item
        raise LookupError(f'Could not find "{menu_item}" menu item')

    def get_menu_items(self, menu_name: str):
        """Return the specified menu."""
        for menu in self.get_menu_bar().get_ax_attribute("AXChildren"):
            if menu.get_ax_attribute("AXTitle") == menu_name:
                return menu.get_ax_attribute("AXChildren")[0].get_ax_attribute("AXChildren")
        raise LookupError(f'Could not find "{menu_name}" menu')

    def get_menu_bar(self):
        """Return app's menu bar."""
        for item in self.application.get_ax_attribute("AXChildren"):
            if item.get_ax_attribute("AXRole") == "AXMenuBar":
                return item
        raise LookupError("Could not find menu bar")

    def activate(self) -> bool:
        """Activate the application."""
        return self._running_app.activateWithOptions_(
            NSApplicationActivationOptions.ActivateIgnoringOtherWindows
        )

    def get_ax_attribute(self, attribute_name: str):
        """Get the value of the the specified attribute."""
        try:
            return self.converter.convert_value(
                get_accessibility_element_attribute(self.ref, attribute_name)
            )
        except (AXErrorNoValue, AXErrorAttributeUnsupported):
            return [] if attribute_name == "AXChildren" else None

    def set_ax_attribute(self, name, value):
        """Set the specified attribute to the specified value."""
        if not check_attribute_settable(self.ref, name):
            raise AXErrorUnsupported(f'Attribute "{name}" is not settable')
        set_attribute_value(self.ref, name, value)

    def press(self):
        self.perform_ax_action("AXPress")

    def perform_ax_action(self, name):
        """Perform specified action on the element."""
        perform_action_on_element(self.ref, name)

    @property
    def ax_actions(self) -> List[str]:
        """Get the list of actions available on the AXUIElement."""
        try:
            return list(get_element_action_names(self.ref))
        except AXError:
            return []

    @property
    def ax_attributes(self) -> List[str]:
        """Get the list of attributes available on the AXUIElement."""
        try:
            return list(get_element_attribute_names(self.ref))
        except AXError:
            return []

    @property
    def localized_name(self):
        """Return the localized name of the application."""
        return self.application.get_ax_attribute("AXTitle")

    @property
    def application(self):
        """Get the application accessibility object."""
        if self.__application is None:
            if self.get_ax_attribute("AXRole") == "AXApplication":
                self.__application = self
            else:
                self.__application = self.get_ax_attribute("AXTopLevelUIElement").get_ax_attribute(
                    "AXParent"
                )
        return self.__application

    def get_element_at_position(self, x: int, y: int):
        """Return the AXUIElement at the given coordinates."""
        if self.ref is None:
            raise AXErrorUnsupported("Operation not supported on null element references")
        element = get_accessibility_object_on_screen_position(self.ref, x, y)
        return self.__class__(element)

    @property
    def bundle_id(self) -> str:
        """Get the AXUIElement's bundle identifier"""
        return self._running_app.bundleIdentifier()

    @property
    def _running_app(self):
        # noinspection PyUnresolvedReferences
        return AppKit.NSRunningApplication.runningApplicationWithProcessIdentifier_(self.pid)

    @property
    def pid(self) -> int:
        """Get the AXUIElement's process ID"""
        return get_accessibility_object_pid(self.ref)

    def set_timeout(self, timeout: int) -> None:
        """Set the timeout value used in the accessibility API."""
        set_accessibility_api_timeout(self.ref, timeout)

    @property
    def focused_window(self):
        """Return the focused application window."""
        return self.application.get_ax_attribute("AXFocusedWindow")

    @property
    def windows(self):
        """Return application windows."""
        return self.application.get_ax_attribute("AXWindows")

    @property
    def children(self) -> list:
        return self.get_ax_attribute("AXChildren")

    def find_element(self, recursive: bool = False, **kwargs):
        """Return the first object that matches lookup criteria."""
        for item in self.get_children(recursive=recursive):
            if match_filter(**kwargs)(item):
                return item

    def find_elements(self, recursive: bool = False, **kwargs):
        """Return a list of all child elements that match lookup criteria."""
        return list(filter(match_filter(**kwargs), self.get_children(recursive=recursive)))

    def get_children(self, target=None, recursive: bool = False):
        """Generator yielding child objects."""
        if target is None:
            target = self
        for child in target.get_ax_attribute("AXChildren"):
            yield child
            if recursive:
                for c in self.get_children(child, recursive):
                    yield c


def match_filter(**attributes):
    def _match(ax_object):
        for attribute in attributes.keys():
            try:
                value = ax_object.get_ax_attribute(attribute)
                if value != attributes.get(attribute):
                    return False
            except AttributeError:
                return False
        return True

    return _match



import re
from collections import namedtuple

from ApplicationServices import AXUIElementGetTypeID
from ApplicationServices import AXValueGetType
from ApplicationServices import NSPointFromString
from ApplicationServices import NSRangeFromString
from ApplicationServices import NSSizeFromString
from ApplicationServices import kAXValueCFRangeType
from ApplicationServices import kAXValueCGPointType
from ApplicationServices import kAXValueCGSizeType
from CoreFoundation import CFArrayGetTypeID
from CoreFoundation import CFGetTypeID
from CoreFoundation import CFStringGetTypeID


class Converter:
    def __init__(self, ax_ui_element_class=None):
        self.app_ref_class = ax_ui_element_class

    def convert_value(self, value):
        if CFGetTypeID(value) == CFStringGetTypeID():
            try:
                return str(value)
            except UnicodeEncodeError:
                return str(value.encode("utf-8"))
        if CFGetTypeID(value) == AXUIElementGetTypeID():
            return self.convert_app_ref(value)
        if CFGetTypeID(value) == CFArrayGetTypeID():
            return self.convert_list(value)
        if AXValueGetType(value) == kAXValueCGSizeType:
            return self.convert_size(value)
        if AXValueGetType(value) == kAXValueCGPointType:
            return self.convert_point(value)
        if AXValueGetType(value) == kAXValueCFRangeType:
            return self.convert_range(value)
        else:
            return value

    def convert_list(self, value):
        return [self.convert_value(item) for item in value]

    def convert_app_ref(self, value):
        return self.app_ref_class(ref=value)

    @staticmethod
    def convert_size(value):
        repr_searched = re.search("{.*}", str(value)).group()
        CGSize = namedtuple("CGSize", ["width", "height"])
        size = NSSizeFromString(repr_searched)
        return CGSize(size.width, size.height)

    @staticmethod
    def convert_point(value):
        repr_searched = re.search("{.*}", str(value)).group()
        CGPoint = namedtuple("CGPoint", ["x", "y"])
        point = NSPointFromString(repr_searched)
        return CGPoint(point.x, point.y)

    @staticmethod
    def convert_range(value):
        repr_searched = re.search("{.*}", str(value)).group()
        CFRange = namedtuple("CFRange", ["location", "length"])
        _range_ = NSRangeFromString(repr_searched)
        return CFRange(_range_.location, _range_.length)



"""Converts types from Python to Apple Event Manager and vice-versa."""

import datetime
import struct
from typing import Any

from Foundation import NSURL
from Foundation import NSAppleEventDescriptor

from . import aeobjects

__all__ = ["AEEnum", "AEType", "ae_converter", "four_characters_code"]


def four_characters_code(code: bytes) -> int:
    """Convert four-char code for use in NSAppleEventDescriptor methods."""
    return struct.unpack(">I", code)[0]


class AEConverter:
    """Implements mappings for common Python types with direct AppleScript equivalents."""

    kMacEpoch = datetime.datetime(1904, 1, 1)
    kUSRF = four_characters_code(aeobjects.keyASUserRecordFields)

    def __init__(self):
        # Clients may add/remove/replace encoder and decoder items:
        self.encoders = {
            NSAppleEventDescriptor.class__(): self.pack_description,
            type(None): self.pack_none,
            bool: self.pack_bool,
            int: self.pack_int,
            float: self.pack_float,
            bytes: self.pack_bytes,
            str: self.pack_str,
            list: self.pack_list,
            tuple: self.pack_list,
            dict: self.pack_dict,
            datetime.datetime: self.pack_datetime,
            AEType: self.pack_type,
            AEEnum: self.pack_enum,
        }

        self.decoders = {
            four_characters_code(k): v
            for k, v in {
                aeobjects.typeNull: self.unpack_null,
                aeobjects.typeBoolean: self.unpack_boolean,
                aeobjects.typeFalse: self.unpack_boolean,
                aeobjects.typeTrue: self.unpack_boolean,
                aeobjects.typeSInt32: self.unpack_s_int32,
                aeobjects.typeIEEE64BitFloatingPoint: self.unpack_float64,
                aeobjects.typeUTF8Text: self.unpack_unicode_text,
                aeobjects.typeUTF16ExternalRepresentation: self.unpack_unicode_text,
                aeobjects.typeUnicodeText: self.unpack_unicode_text,
                aeobjects.typeLongDateTime: self.unpack_long_datetime,
                aeobjects.typeAEList: self.unpack_ae_list,
                aeobjects.typeAERecord: self.unpack_ae_record,
                aeobjects.typeAlias: self.unpack_file,
                aeobjects.typeFSS: self.unpack_file,
                aeobjects.typeFSRef: self.unpack_file,
                aeobjects.typeFileURL: self.unpack_file,
                aeobjects.typeType: self.unpack_type,
                aeobjects.typeEnumeration: self.unpack_enumeration,
            }.items()
        }

    def pack(self, data: Any) -> NSAppleEventDescriptor:
        """Pack Python data.
        Result : NSAppleEventDescriptor -- an AE descriptor,
        or error if no encoder exists for this type_ of data
        """
        try:
            return self.encoders[data.__class__](data)  # quick lookup by type_/class
        except (KeyError, AttributeError):
            for type_, encoder in self.encoders.items():
                # slower but more thorough lookup that can handle subtypes/subclasses
                if isinstance(data, type_):
                    return encoder(data)
        raise TypeError("Can't pack data into an AEDesc (unsupported type_): {!r}".format(data))

    def unpack(self, desc: NSAppleEventDescriptor) -> Any:
        """Unpack Apple event descriptor.
        Returns Python value or the original NSAppleEventDescriptor if no decoder is found.
        """
        decoder = self.decoders.get(desc.descriptorType())
        if decoder:
            # Unpack known type.
            return decoder(desc)
        # If it is a record-like descriptor, unpack as dict
        # with an extra AEType(b'pcls') key containing the descriptor type.
        rec = desc.coerceToDescriptorType_(four_characters_code(aeobjects.typeAERecord))
        if rec:
            rec = self.unpack_ae_record(rec)
            rec[AEType(aeobjects.pClass)] = AEType(struct.pack(">I", desc.descriptorType()))
            return rec
        # Return descriptor as-is.
        return desc

    @staticmethod
    def _pack_bytes(description_type, data):
        return NSAppleEventDescriptor.descriptorWithDescriptorType_bytes_length_(
            four_characters_code(description_type), data, len(data)
        )

    @staticmethod
    def pack_description(val):
        return val

    @staticmethod
    def pack_none():
        return NSAppleEventDescriptor.nullDescriptor()

    @staticmethod
    def pack_bool(val):
        return NSAppleEventDescriptor.descriptorWithBoolean_(int(val))

    def pack_int(self, val):
        if (-(2 ** 31)) <= val < (2 ** 31):
            return NSAppleEventDescriptor.descriptorWithInt32_(val)
        else:
            return self.pack(float(val))

    def pack_float(self, val):
        return self._pack_bytes(aeobjects.typeFloat, struct.pack("d", val))

    def pack_bytes(self, val):
        return self._pack_bytes(aeobjects.typeData, val)

    @staticmethod
    def pack_str(val):
        return NSAppleEventDescriptor.descriptorWithString_(val)

    def pack_datetime(self, val):
        delta = val - self.kMacEpoch
        sec = delta.days * 3600 * 24 + delta.seconds
        return self._pack_bytes(aeobjects.typeLongDateTime, struct.pack("q", sec))

    def pack_list(self, val):
        lst = NSAppleEventDescriptor.listDescriptor()
        for item in val:
            lst.insertDescriptor_atIndex_(self.pack(item), 0)
        return lst

    def pack_dict(self, val):
        record = NSAppleEventDescriptor.recordDescriptor()
        usrf = desctype = None
        for key, value in val.items():
            if isinstance(key, AEType):
                if key.code == aeobjects.pClass and isinstance(value, AEType):
                    # AS packs records that contain a 'class' property by coercing the packed record
                    # to the descriptor type specified by a property value (assuming it's an AEType)
                    desctype = value
                else:
                    record.setDescriptor_forKeyword_(
                        self.pack(value), four_characters_code(key.code)
                    )
            else:
                if not usrf:
                    usrf = NSAppleEventDescriptor.listDescriptor()
                usrf.insertDescriptor_atIndex_(self.pack(key), 0)
                usrf.insertDescriptor_atIndex_(self.pack(value), 0)
        if usrf:
            record.setDescriptor_forKeyword_(usrf, self.kUSRF)
        if desctype:
            new_record = record.coerceToDescriptorType_(four_characters_code(desctype.code))
            if new_record:
                record = new_record
            else:  # coercion failed for some reason, so pack as normal key-value pair
                # noinspection PyUnboundLocalVariable
                record.setDescriptor_forKeyword_(
                    self.pack(desctype), four_characters_code(key.code)
                )
        return record

    @staticmethod
    def pack_type(val):
        return NSAppleEventDescriptor.descriptorWithTypeCode_(four_characters_code(val.code))

    @staticmethod
    def pack_enum(val):
        return NSAppleEventDescriptor.descriptorWithEnumCode_(four_characters_code(val.code))

    @staticmethod
    def unpack_null(desc):
        _ = desc
        return None

    @staticmethod
    def unpack_boolean(desc):
        return desc.booleanValue()

    @staticmethod
    def unpack_s_int32(desc):
        return desc.int32Value()

    @staticmethod
    def unpack_float64(desc):
        return struct.unpack("d", bytes(desc.data()))[0]

    @staticmethod
    def unpack_unicode_text(desc):
        return desc.stringValue()

    def unpack_long_datetime(self, desc):
        return self.kMacEpoch + datetime.timedelta(
            seconds=struct.unpack("q", bytes(desc.data()))[0]
        )

    def unpack_ae_list(self, desc):
        return [self.unpack(desc.descriptorAtIndex_(i + 1)) for i in range(desc.numberOfItems())]

    def unpack_ae_record(self, desc):
        dct = {}
        for i in range(desc.numberOfItems()):
            key = desc.keywordForDescriptorAtIndex_(i + 1)
            value = desc.descriptorForKeyword_(key)
            if key == self.kUSRF:
                lst = self.unpack_ae_list(value)
                # noinspection PyAssignmentToLoopOrWithParameter
                for i in range(0, len(lst), 2):
                    dct[lst[i]] = lst[i + 1]
            else:
                dct[AEType(struct.pack(">I", key))] = self.unpack(value)
        return dct

    @staticmethod
    def unpack_type(desc):
        return AEType(struct.pack(">I", desc.typeCodeValue()))

    @staticmethod
    def unpack_enumeration(desc):
        return AEEnum(struct.pack(">I", desc.enumCodeValue()))

    @staticmethod
    def unpack_file(desc):
        url = bytes(
            desc.coerceToDescriptorType_(four_characters_code(aeobjects.typeFileURL)).data()
        ).decode("utf8")
        return NSURL.URLWithString_(url).path()


class AETypeBase:
    """Base class for AEType and AEEnum.
    Hashable and comparable, so may be used as keys in dictionaries that map to AE records.
    """

    def __init__(self, code: bytes):
        if not isinstance(code, bytes):
            raise TypeError("invalid code (not a bytes object): {!r}".format(code))
        elif len(code) != 4:
            raise ValueError("invalid code (not four bytes long): {!r}".format(code))
        self._code = code

    code = property(lambda self: self._code, doc="bytes -- four-char code, e.g. b'utxt'")

    def __hash__(self):
        return hash(self._code)

    def __eq__(self, val):
        return val.__class__ == self.__class__ and val.code == self._code

    def __ne__(self, val):
        return not self == val

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self._code)


class AEType(AETypeBase):
    """AE type. Maps to an AppleScript type class, e.g. AEType(b'utxt') <=> 'unicode text'."""


class AEEnum(AETypeBase):
    """AE enumeration. Maps to an AppleScript constant, e.g. AEEnum(b'yes ') <=> 'yes'."""


ae_converter = AEConverter()



from typing import ClassVar
from typing import Optional
from typing import Tuple

from macuitest.lib.core import slow_down
from macuitest.lib.operating_system.env import env


class Defaults:
    preferences_daemon: ClassVar[str] = "cfprefsd"

    def __init__(self, executor, service_manager):
        self.executor = executor
        self.service_manager = service_manager

    def clear_application_preferences(self, bundle_ids: Tuple[str, ...]):
        """Clear application preferences.
        :param bundle_ids: Tuple of target bundle ids."""
        for bundle_id in bundle_ids:
            self.remove_property_list_by_id(bundle_id, flush_pref_cache=False)
            self.executor.execute(f"rm -f {env.user_preferences}/{bundle_id}*")

    def remove_property_list_by_id(self, bundle_id: str, flush_pref_cache: bool = True):
        self.delete_defaults(bundle_id)
        if flush_pref_cache:
            self.flush_preferences_cache()

    def read_defaults(
        self, domain: str = "NSGlobalDomain", key: Optional[str] = None
    ) -> Optional[str]:
        command: str = f"defaults read {domain} {key}" if key else f"defaults read {domain}"
        response = self.executor.execute(command)
        if response.returncode == 0:
            return response.stdout.decode(encoding="utf-8").strip()

    def delete_defaults(self, domain: str = "NSGlobalDomain", key: Optional[str] = None) -> int:
        command: str = f"defaults delete {domain} {key}" if key else f"defaults delete {domain}"
        return self.executor.execute(command.strip()).returncode

    @slow_down(seconds=0.5)
    def write_defaults(
        self, domain: str = "NSGlobalDomain", key: str = "KeyRepeat", value="0.02"
    ) -> int:
        return self.executor.execute(f"defaults write {domain} {key} {value}").returncode

    @slow_down(seconds=2)
    def flush_preferences_cache(self):
        self.service_manager.kill_process(self.preferences_daemon)



import os
import warnings
from pathlib import Path
from typing import ClassVar

import cv2
import pytesseract

from macuitest.config.constants import Region
from macuitest.lib.core import wait_condition
from macuitest.lib.elements.ui.monitor import monitor


class OCRManager:
    ocr_engine_mode: ClassVar[int] = 3
    page_segmentation_mode: ClassVar[int] = 6
    tesseract_config: ClassVar[str] = f"--oem {ocr_engine_mode} --psm {page_segmentation_mode}"

    def __init__(self, language: str = "eng"):
        if os.system("tesseract --help > /dev/null 2>&1") != 0:
            warnings.warn("Tesseract is not installed, please install: brew install tesseract")
        self.language = language
        self.trained_data = f"{self.language}.traineddata"
        self.trained_data_path: Path = Path("/usr/local/share/tessdata")
        self.__assert_trained_data_present()

    def wait_text(self, text: str, where: Region, timeout: int = 10) -> bool:
        return wait_condition(lambda: self.recognize(region=where) == text, timeout=timeout)

    def recognize(self, region: Region, is_font_white: bool = False) -> str:
        img_gray = cv2.cvtColor(monitor.make_snapshot(region), cv2.COLOR_BGR2GRAY)
        if (
            is_font_white
        ):  # We want to invert font color to get better character recognition results.
            img_gray = cv2.bitwise_not(img_gray)
        payload = pytesseract.image_to_string(
            img_gray, config=self.tesseract_config, lang=self.language
        )
        return os.linesep.join([s for s in payload.splitlines() if s])

    def __assert_trained_data_present(self):
        if not self.trained_data_path.joinpath(self.trained_data).exists():
            Path(self.trained_data_path).mkdir(parents=True, exist_ok=True)
            raise FileNotFoundError(
                f"Download trained data from: "
                f"https://github.com/tesseract-ocr/tessdata_best/raw/master/{self.trained_data}"
                f"and save it to: {self.trained_data_path.joinpath(self.trained_data)}"
            )


ocr_manager = OCRManager()



import os
from pathlib import Path
from typing import Union

from macuitest.lib.operating_system.env import env


class ScreenshotPathBuilder:
    """Screenshot path builder."""

    def __init__(
        self,
        category: str,
        root: Union[Path, str] = os.environ.get("MACUITEST_SCR", os.environ.get("HOME")),
    ):
        self.root = Path(root)
        self.category = category.lower().replace(" ", "_").replace("-", "_")

    def __getattr__(self, item: str):
        return getattr(self, item) if item == "category" else self.build_path(self.category, item)

    def build_path(self, section: str, scr_name: str) -> str:
        """Build absolute path to a screenshot."""
        base = self.root.joinpath(section).joinpath(f"{scr_name}.png")
        os_ver = env.version[1] if env.version < (11, 0) else env.version[0]
        os_specific = self.root.joinpath(section).joinpath(f"{scr_name}_{os_ver}.png")
        return (os_specific if os_specific.exists() else base).as_posix()



from dataclasses import dataclass
from types import MappingProxyType
from typing import Dict
from typing import Tuple
from typing import Union

KB: int = 10 ** 3
MB: int = 10 ** 6
GB: int = 10 ** 9

SECOND: int = 1
MINUTE: int = 60 * SECOND
HOUR: int = 60 * MINUTE


class DNSMapping(str):
    pass


@dataclass(frozen=True)
class Point:
    x: Union[int, float]
    y: Union[int, float]


@dataclass(frozen=True)
class Frame:
    x1: Union[int, float]
    y1: Union[int, float]
    x2: Union[int, float]
    y2: Union[int, float]
    center: Point
    width: Union[int, float]
    height: Union[int, float]


@dataclass
class Region:
    x1: Union[int, float]
    y1: Union[int, float]
    x2: Union[int, float]
    y2: Union[int, float]


@dataclass(frozen=True)
class ScreenSize:
    width: Union[int, float]
    height: Union[int, float]


@dataclass(frozen=True)
class CheckboxState:
    states: MappingProxyType = MappingProxyType(
        {"unchecked": 0, 0: 0, "checked": 1, 1: 1, "indeterminate": 2, 2: 2}
    )
    checked: int = 1
    unchecked: int = 0
    indeterminate: int = 2


@dataclass(frozen=True)
class Colors:
    red: Tuple[str, ...] = (
        "lightcoral",
        "lightsalmon",
        "salmon",
        "tomato",
    )
    green: Tuple[str, ...] = (
        "lightgreen",
        "limegreen",
        "mediumaquamarine",
        "mediumseagreen",
        "darkseagreen",
    )
    blue: Tuple[str, ...] = (
        "deepskyblue",
        "dodgerblue",
        "cornflowerblue",
        "royalblue",
        "lavender",
        "lightskyblue",
        "mediumturquoise",
    )
    orange: Tuple[str, ...] = (
        "coral",
        "lightsalmon",
        "sandybrown",
        "darkorange",
        "goldenrod",
    )
    grey: Tuple[str, ...] = (
        "darkgray",
        "darkslategray",
        "dimgray",
        "lightgray",
        "slategray",
        "darkgrey",
        "darkslategrey",
        "dimgrey",
        "lightgrey",
        "slategrey",
        "silver",
    )


@dataclass(frozen=True)
class DisclosureTriangleState:
    closed: int = 0
    opened: int = 1


@dataclass
class URLDetails:
    url: str
    schema: str
    domain: str
    path: str
    query: Dict[str, str]



"""Wrap objc calls to raise python exception."""
import AppKit
from ApplicationServices import AXIsProcessTrusted
from ApplicationServices import AXUIElementCopyActionNames
from ApplicationServices import AXUIElementCopyAttributeNames
from ApplicationServices import AXUIElementCopyAttributeValue
from ApplicationServices import AXUIElementCopyElementAtPosition
from ApplicationServices import AXUIElementGetPid
from ApplicationServices import AXUIElementIsAttributeSettable
from ApplicationServices import AXUIElementPerformAction
from ApplicationServices import AXUIElementSetAttributeValue
from ApplicationServices import AXUIElementSetMessagingTimeout
from ApplicationServices import kAXErrorActionUnsupported
from ApplicationServices import kAXErrorAPIDisabled
from ApplicationServices import kAXErrorAttributeUnsupported
from ApplicationServices import kAXErrorCannotComplete
from ApplicationServices import kAXErrorFailure
from ApplicationServices import kAXErrorIllegalArgument
from ApplicationServices import kAXErrorInvalidUIElement
from ApplicationServices import kAXErrorInvalidUIElementObserver
from ApplicationServices import kAXErrorNotificationAlreadyRegistered
from ApplicationServices import kAXErrorNotificationNotRegistered
from ApplicationServices import kAXErrorNotificationUnsupported
from ApplicationServices import kAXErrorNotImplemented
from ApplicationServices import kAXErrorNoValue
from ApplicationServices import kAXErrorSuccess
from PyObjCTools import AppHelper


def is_accessibility_enabled():
    """Return the status of accessibility on the system."""
    return AXIsProcessTrusted()


def get_frontmost_pid():
    """Return the process ID of the application in the foreground"""
    return AppKit.NSWorkspace.sharedWorkspace().frontmostApplication().processIdentifier()


def get_running_apps():
    """Get a list of the running applications"""
    AppHelper.callLater(1, AppHelper.stopEventLoop)
    AppHelper.runConsoleEventLoop()
    work_space = AppKit.NSWorkspace.sharedWorkspace()
    return work_space.runningApplications()


def launch_app_by_bundle_id(bundle_id: str):
    work_space = AppKit.NSWorkspace.sharedWorkspace()
    response = work_space.launchAppWithBundleIdentifier_options_additionalEventParamDescriptor_launchIdentifier_(  # noqa: E501
        bundle_id,
        AppKit.NSWorkspaceLaunchAllowingClassicStartup,
        AppKit.NSAppleEventDescriptor.nullDescriptor(),
        None,
    )
    if not response[0]:
        raise RuntimeError(f"Error launching specified application. {response}")


def launch_app_by_bundle_path(bundle_path: str, arguments=None):
    if arguments is None:
        arguments = []
    bundle_url = AppKit.NSURL.fileURLWithPath_(bundle_path)
    workspace = AppKit.NSWorkspace.sharedWorkspace()
    configuration = {AppKit.NSWorkspaceLaunchConfigurationArguments: arguments}
    return workspace.launchApplicationAtURL_options_configuration_error_(
        bundle_url, AppKit.NSWorkspaceLaunchAllowingClassicStartup, configuration, None
    )


def terminate_app_by_bundle_id(bundle_id: str):
    apps = get_running_apps_with_bundle_id(bundle_id)
    if apps:
        apps[0].terminate()


def get_running_apps_with_bundle_id(bundle_id: str):
    """Return an array of NSRunningApplications."""
    return AppKit.NSRunningApplication.runningApplicationsWithBundleIdentifier_(bundle_id)


def get_accessibility_element_attribute(element, attribute):
    """Return the value of an accessibility object's attribute
    Args:
        element: The AXUIElementRef representing the accessibility object
        attribute: The attribute name

    Returns: the value associated with the specified attribute
    """
    error_code, attr_value = AXUIElementCopyAttributeValue(element, attribute, None)
    error_messages = {
        kAXErrorAttributeUnsupported: "The specified AXUIElementRef "
        "does not support the specified attribute.",
        kAXErrorNoValue: "The specified attribute does not have a value.",
        kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        kAXErrorCannotComplete: "The function cannot complete "
        "because messaging has failed in some way.",
        kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    check_ax_error(error_code, error_messages)
    return attr_value


def check_attribute_settable(element, attribute) -> bool:
    """Check whether the specified accessibility object's attribute can be modified.
    Args:
        element: The AXUIElementRef representing the accessibility object
        attribute: The attribute name

    Returns: a Boolean value indicating whether the attribute is settable
    """
    error_code, settable = AXUIElementIsAttributeSettable(element, attribute, None)
    error_messages = {
        kAXErrorCannotComplete: "The function cannot complete because messaging "
        "has failed in some way (often due to a timeout).",
        kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
        kAXErrorAttributeUnsupported: "The specified AXUIElementRef "
        "does not support the specified attribute.",
        kAXErrorNoValue: "The specified attribute does not have a value.",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    check_ax_error(error_code, error_messages)
    return settable


def set_attribute_value(element, attribute, value):
    """Set the accessibility object's attribute to the specified value
    Args:
        element: The AXUIElementRef representing the accessibility object
        attribute: The attribute name
        value: The new value for the attribute
    """
    error_code = AXUIElementSetAttributeValue(element, attribute, value)
    error_messages = {
        kAXErrorIllegalArgument: "The value is not recognized by the accessible application"
        " or one of the other arguments is an illegal value.",
        kAXErrorAttributeUnsupported: "The specified AXUIElementRef "
        "does not support the specified attribute.",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        kAXErrorCannotComplete: "The function cannot complete "
        "because messaging has failed in some way.",
        kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    check_ax_error(error_code, error_messages)


def get_element_attribute_names(element):
    """Get a list of attributes supported by the specified accessibility object
    Args:
        element: The AXUIElementRef representing the accessibility object

    Returns: an array containing the accessibility object's attribute names
    """
    error_code, names = AXUIElementCopyAttributeNames(element, None)
    error_messages = {
        kAXErrorAttributeUnsupported: "The specified AXUIElementRef "
        "does not support the specified attribute.",
        kAXErrorIllegalArgument: "One or both of the arguments is an illegal value.",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        kAXErrorFailure: "There was a system memory failure.",
        kAXErrorCannotComplete: "The function cannot complete "
        "because messaging has failed in some way.",
        kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    check_ax_error(error_code, error_messages)
    return names


def get_element_action_names(element):
    """Get a list of all the actions the specified accessibility object can perform.
    Args:
        element: The AXUIElementRef representing the accessibility object

    Returns: an array of actions the accessibility object can perform
        (empty if the accessibility object supports no actions)
    """
    error_code, names = AXUIElementCopyActionNames(element, None)
    error_messages = {
        kAXErrorIllegalArgument: "One or both of the arguments is an illegal value.",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        kAXErrorFailure: "There was some sort of system memory failure.",
        kAXErrorCannotComplete: "The function cannot complete "
        "because messaging has failed in some way.",
        kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    check_ax_error(error_code, error_messages)
    return names


def perform_action_on_element(element, action):
    """Make the accessibility object perform the specified action.
    Args:
        element: The AXUIElementRef representing the accessibility object
        action: The action to be performed
    """
    error_code = AXUIElementPerformAction(element, action)
    error_messages = {
        kAXErrorActionUnsupported: "Specified AXUIElementRef does not support the specified action"
        " (you will also receive this error error "
        "if you pass in the system-wide accessibility object).",
        kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        kAXErrorCannotComplete: "The function cannot complete because messaging has failed"
        " in some way or the application has not yet responded.",
        kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    check_ax_error(error_code, error_messages)


def get_accessibility_object_pid(element):
    """Get the process ID associated with the specified accessibility object.
    Args:
        element: The AXUIElementRef representing an accessibility object

    Returns: the process ID associated with the specified accessibility object
    """
    error_code, pid = AXUIElementGetPid(element, None)
    error_messages = {
        kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
    }
    check_ax_error(error_code, error_messages)
    return pid


def get_accessibility_object_on_screen_position(application, x, y):
    """Get the accessibility object at the specified position
    in top-left relative screen coordinates
    Args:
        application: The AXUIElementRef representing the application that
            contains the screen coordinates (or the system-wide accessibility object)
        x: The horizontal position
        y: The vertical position

    Returns: the accessibility object at the position specified by x and y
    """
    error_code, element = AXUIElementCopyElementAtPosition(application, x, y, None)
    error_messages = {
        kAXErrorNoValue: "There is no accessibility object at the specified position.",
        kAXErrorIllegalArgument: "One or more of the arguments is an illegal value.",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
        kAXErrorCannotComplete: "The function cannot complete "
        "because messaging has failed in some way.",
        kAXErrorNotImplemented: "The process does not fully support the accessibility API.",
    }
    check_ax_error(error_code, error_messages)
    return element


def set_accessibility_api_timeout(element, timeout: int):
    """Set the timeout value used in the accessibility API
    Args:
        element: The AXUIElementRef representing an accessibility object
        timeout: The number of seconds for the new timeout value
    """
    error_code = AXUIElementSetMessagingTimeout(element, timeout)
    error_messages = {
        kAXErrorIllegalArgument: "One or more of the arguments "
        "is an illegal value (timeout values must be positive).",
        kAXErrorInvalidUIElement: "The AXUIElementRef is invalid.",
    }
    check_ax_error(error_code, error_messages)


class AXError(Exception):
    pass


class AXErrorUnsupported(AXError):
    pass


class AXErrorAPIDisabled(AXError):
    pass


class AXErrorInvalidUIElement(AXError):
    pass


class AXErrorCannotComplete(AXError):
    pass


class AXErrorNotImplemented(AXError):
    pass


class AXErrorIllegalArgument(AXError):
    pass


class AXErrorActionUnsupported(AXError):
    pass


class AXErrorNoValue(AXError):
    pass


class AXErrorFailure(AXError):
    pass


class AXErrorInvalidUIElementObserver(AXError):
    pass


class AXErrorNotificationUnsupported(AXError):
    pass


class AXErrorNotificationAlreadyRegistered(AXError):
    pass


class AXErrorNotificationNotRegistered(AXError):
    pass


class AXErrorAttributeUnsupported(AXError):
    pass


class AXErrorFactory:
    def __new__(cls, error_code):
        return {
            kAXErrorAPIDisabled: AXErrorAPIDisabled,
            kAXErrorInvalidUIElement: AXErrorInvalidUIElement,
            kAXErrorCannotComplete: AXErrorCannotComplete,
            kAXErrorNotImplemented: AXErrorNotImplemented,
            kAXErrorIllegalArgument: AXErrorIllegalArgument,
            kAXErrorNoValue: AXErrorNoValue,
            kAXErrorFailure: AXErrorFailure,
            kAXErrorInvalidUIElementObserver: AXErrorInvalidUIElementObserver,
            kAXErrorNotificationUnsupported: AXErrorNotificationUnsupported,
            kAXErrorNotificationAlreadyRegistered: AXErrorNotificationAlreadyRegistered,
            kAXErrorNotificationNotRegistered: AXErrorNotificationNotRegistered,
            kAXErrorAttributeUnsupported: AXErrorAttributeUnsupported,
            kAXErrorActionUnsupported: AXErrorActionUnsupported,
        }.get(error_code, AXErrorUnsupported)


def check_ax_error(error_code: int, error_messages) -> None:
    """Return if code is kAXErrorSuccess.
    Raise an error with given message based on given error code.
    Defaults to AXErrorUnsupported for unknown codes.
       Args:
           error_code: the error code
           error_messages: mapping from error code to error message
    """
    if error_code == kAXErrorSuccess:
        return
    try:
        error_message = error_messages[error_code]
    except KeyError:
        if error_code == kAXErrorFailure:
            error_message = "There is some sort of system memory failure"
        else:
            error_message = f"Unknown AX Error code: {error_code}"
    raise AXErrorFactory(error_code)(error_message)



import string

KEY_NAMES = [
    "\t",
    "\n",
    "\r",
    " ",
    "!",
    '"',
    "#",
    "$",
    "%",
    "&",
    "'",
    "(",
    ")",
    "*",
    "+",
    ",",
    "-",
    ".",
    "/",
    "0",
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
    ":",
    ";",
    "<",
    "=",
    ">",
    "?",
    "@",
    "[",
    "\\",
    "]",
    "^",
    "_",
    "`",
    "a",
    "b",
    "c",
    "d",
    "e",
    "f",
    "g",
    "h",
    "i",
    "j",
    "k",
    "l",
    "m",
    "n",
    "o",
    "p",
    "q",
    "r",
    "s",
    "t",
    "u",
    "v",
    "w",
    "x",
    "y",
    "z",
    "{",
    "|",
    "}",
    "~",
    "accept",
    "add",
    "alt",
    "altleft",
    "altright",
    "apps",
    "backspace",
    "browserback",
    "browserfavorites",
    "browserforward",
    "browserhome",
    "browserrefresh",
    "browsersearch",
    "browserstop",
    "capslock",
    "clear",
    "convert",
    "ctrl",
    "ctrlleft",
    "ctrlright",
    "decimal",
    "del",
    "delete",
    "divide",
    "down",
    "end",
    "enter",
    "esc",
    "escape",
    "execute",
    "f1",
    "f10",
    "f11",
    "f12",
    "f13",
    "f14",
    "f15",
    "f16",
    "f17",
    "f18",
    "f19",
    "f2",
    "f20",
    "f21",
    "f22",
    "f23",
    "f24",
    "f3",
    "f4",
    "f5",
    "f6",
    "f7",
    "f8",
    "f9",
    "final",
    "fn",
    "hanguel",
    "hangul",
    "hanja",
    "help",
    "home",
    "insert",
    "junja",
    "kana",
    "kanji",
    "launchapp1",
    "launchapp2",
    "launchmail",
    "launchmediaselect",
    "left",
    "modechange",
    "multiply",
    "nexttrack",
    "nonconvert",
    "num0",
    "num1",
    "num2",
    "num3",
    "num4",
    "num5",
    "num6",
    "num7",
    "num8",
    "num9",
    "numlock",
    "pagedown",
    "pageup",
    "pause",
    "pgdn",
    "pgup",
    "playpause",
    "prevtrack",
    "print",
    "printscreen",
    "prntscrn",
    "prtsc",
    "prtscr",
    "return",
    "right",
    "scrolllock",
    "select",
    "separator",
    "shift",
    "shiftleft",
    "shiftright",
    "sleep",
    "space",
    "stop",
    "subtract",
    "tab",
    "up",
    "volumedown",
    "volumemute",
    "volumeup",
    "win",
    "winleft",
    "winright",
    "yen",
    "command",
    "option",
    "optionleft",
    "optionright",
]

KEYBOARD_KEYS = dict([(key, None) for key in KEY_NAMES])

""" Taken from events.h
/System/Library/Frameworks/Carbon.framework/Versions/A/Frameworks/HIToolbox.framework/Versions/A/Headers/Events.h

The *KB dictionaries in a string that can be passed to keyDown(),
keyUp(), or press() into the code used for the OS-specific keyboard function.

They should always be lowercase, and the same keys should be used across all OSes."""

KEYBOARD_KEYS.update(
    {
        "a": 0x00,  # kVK_ANSI_A
        "s": 0x01,  # kVK_ANSI_S
        "d": 0x02,  # kVK_ANSI_D
        "f": 0x03,  # kVK_ANSI_F
        "h": 0x04,  # kVK_ANSI_H
        "g": 0x05,  # kVK_ANSI_G
        "z": 0x06,  # kVK_ANSI_Z
        "x": 0x07,  # kVK_ANSI_X
        "c": 0x08,  # kVK_ANSI_C
        "v": 0x09,  # kVK_ANSI_V
        "b": 0x0B,  # kVK_ANSI_B
        "q": 0x0C,  # kVK_ANSI_Q
        "w": 0x0D,  # kVK_ANSI_W
        "e": 0x0E,  # kVK_ANSI_E
        "r": 0x0F,  # kVK_ANSI_R
        "y": 0x10,  # kVK_ANSI_Y
        "t": 0x11,  # kVK_ANSI_T
        "1": 0x12,  # kVK_ANSI_1
        "!": 0x12,  # kVK_ANSI_1
        "2": 0x13,  # kVK_ANSI_2
        "@": 0x13,  # kVK_ANSI_2
        "3": 0x14,  # kVK_ANSI_3
        "#": 0x14,  # kVK_ANSI_3
        "4": 0x15,  # kVK_ANSI_4
        "$": 0x15,  # kVK_ANSI_4
        "6": 0x16,  # kVK_ANSI_6
        "^": 0x16,  # kVK_ANSI_6
        "5": 0x17,  # kVK_ANSI_5
        "%": 0x17,  # kVK_ANSI_5
        "=": 0x18,  # kVK_ANSI_Equal
        "+": 0x18,  # kVK_ANSI_Equal
        "9": 0x19,  # kVK_ANSI_9
        "(": 0x19,  # kVK_ANSI_9
        "7": 0x1A,  # kVK_ANSI_7
        "&": 0x1A,  # kVK_ANSI_7
        "-": 0x1B,  # kVK_ANSI_Minus
        "_": 0x1B,  # kVK_ANSI_Minus
        "8": 0x1C,  # kVK_ANSI_8
        "*": 0x1C,  # kVK_ANSI_8
        "0": 0x1D,  # kVK_ANSI_0
        ")": 0x1D,  # kVK_ANSI_0
        "]": 0x1E,  # kVK_ANSI_RightBracket
        "}": 0x1E,  # kVK_ANSI_RightBracket
        "o": 0x1F,  # kVK_ANSI_O
        "u": 0x20,  # kVK_ANSI_U
        "[": 0x21,  # kVK_ANSI_LeftBracket
        "{": 0x21,  # kVK_ANSI_LeftBracket
        "i": 0x22,  # kVK_ANSI_I
        "p": 0x23,  # kVK_ANSI_P
        "l": 0x25,  # kVK_ANSI_L
        "j": 0x26,  # kVK_ANSI_J
        "'": 0x27,  # kVK_ANSI_Quote
        '"': 0x27,  # kVK_ANSI_Quote
        "k": 0x28,  # kVK_ANSI_K
        ";": 0x29,  # kVK_ANSI_Semicolon
        ":": 0x29,  # kVK_ANSI_Semicolon
        "\\": 0x2A,  # kVK_ANSI_Backslash
        "|": 0x2A,  # kVK_ANSI_Backslash
        ",": 0x2B,  # kVK_ANSI_Comma
        "<": 0x2B,  # kVK_ANSI_Comma
        "/": 0x2C,  # kVK_ANSI_Slash
        "?": 0x2C,  # kVK_ANSI_Slash
        "n": 0x2D,  # kVK_ANSI_N
        "m": 0x2E,  # kVK_ANSI_M
        ".": 0x2F,  # kVK_ANSI_Period
        ">": 0x2F,  # kVK_ANSI_Period
        "`": 0x32,  # kVK_ANSI_Grave
        "~": 0x32,  # kVK_ANSI_Grave
        " ": 0x31,  # kVK_Space
        "space": 0x31,
        "\r": 0x24,  # kVK_Return
        "\n": 0x24,  # kVK_Return
        "enter": 0x24,  # kVK_Return
        "return": 0x24,  # kVK_Return
        "\t": 0x30,  # kVK_Tab
        "tab": 0x30,  # kVK_Tab
        "backspace": 0x33,  # kVK_Delete, which is "Backspace" on OS X.
        "\b": 0x33,  # kVK_Delete, which is "Backspace" on OS X.
        "esc": 0x35,  # kVK_Escape
        "escape": 0x35,  # kVK_Escape
        "command": 0x37,  # kVK_Command
        "shift": 0x38,  # kVK_Shift
        "shiftleft": 0x38,  # kVK_Shift
        "capslock": 0x39,  # kVK_CapsLock
        "option": 0x3A,  # kVK_Option
        "optionleft": 0x3A,  # kVK_Option
        "alt": 0x3A,  # kVK_Option
        "altleft": 0x3A,  # kVK_Option
        "ctrl": 0x3B,  # kVK_Control
        "ctrlleft": 0x3B,  # kVK_Control
        "shiftright": 0x3C,  # kVK_RightShift
        "optionright": 0x3D,  # kVK_RightOption
        "ctrlright": 0x3E,  # kVK_RightControl
        "fn": 0x3F,  # kVK_Function
        "f17": 0x40,  # kVK_F17
        "volumeup": 0x48,  # kVK_VolumeUp
        "volumedown": 0x49,  # kVK_VolumeDown
        "volumemute": 0x4A,  # kVK_Mute
        "f18": 0x4F,  # kVK_F18
        "f19": 0x50,  # kVK_F19
        "f20": 0x5A,  # kVK_F20
        "f5": 0x60,  # kVK_F5
        "f6": 0x61,  # kVK_F6
        "f7": 0x62,  # kVK_F7
        "f3": 0x63,  # kVK_F3
        "f8": 0x64,  # kVK_F8
        "f9": 0x65,  # kVK_F9
        "f11": 0x67,  # kVK_F11
        "f13": 0x69,  # kVK_F13
        "f16": 0x6A,  # kVK_F16
        "f14": 0x6B,  # kVK_F14
        "f10": 0x6D,  # kVK_F10
        "f12": 0x6F,  # kVK_F12
        "f15": 0x71,  # kVK_F15
        "help": 0x72,  # kVK_Help
        "home": 0x73,  # kVK_Home
        "pageup": 0x74,  # kVK_PageUp
        "pgup": 0x74,  # kVK_PageUp
        "del": 0x75,  # kVK_ForwardDelete
        "delete": 0x75,  # kVK_ForwardDelete
        "f4": 0x76,  # kVK_F4
        "end": 0x77,  # kVK_End
        "f2": 0x78,  # kVK_F2
        "pagedown": 0x79,  # kVK_PageDown
        "pgdn": 0x79,  # kVK_PageDown
        "f1": 0x7A,  # kVK_F1
        "left": 0x7B,  # kVK_LeftArrow
        "right": 0x7C,  # kVK_RightArrow
        "down": 0x7D,  # kVK_DownArrow
        "up": 0x7E,  # kVK_UpArrow
    }
)

for char in string.ascii_lowercase:  # Add mappings for uppercase letters.
    KEYBOARD_KEYS[char.upper()] = KEYBOARD_KEYS[char]

# Taken from ev_keymap.h
# http://www.opensource.apple.com/source/IOHIDFamily/IOHIDFamily-86.1/IOHIDSystem/IOKit/hidsystem/ev_keymap.h
SPECIAL_KEYS = {
    "KEYTYPE_SOUND_UP": 0,
    "KEYTYPE_SOUND_DOWN": 1,
    "KEYTYPE_BRIGHTNESS_UP": 2,
    "KEYTYPE_BRIGHTNESS_DOWN": 3,
    "KEYTYPE_CAPS_LOCK": 4,
    "KEYTYPE_HELP": 5,
    "POWER_KEY": 6,
    "KEYTYPE_MUTE": 7,
    "UP_ARROW_KEY": 8,
    "DOWN_ARROW_KEY": 9,
    "KEYTYPE_NUM_LOCK": 10,
    "KEYTYPE_CONTRAST_UP": 11,
    "KEYTYPE_CONTRAST_DOWN": 12,
    "KEYTYPE_LAUNCH_PANEL": 13,
    "KEYTYPE_EJECT": 14,
    "KEYTYPE_VIDMIRROR": 15,
    "KEYTYPE_PLAY": 16,
    "KEYTYPE_NEXT": 17,
    "KEYTYPE_PREVIOUS": 18,
    "KEYTYPE_FAST": 19,
    "KEYTYPE_REWIND": 20,
    "KEYTYPE_ILLUMINATION_UP": 21,
    "KEYTYPE_ILLUMINATION_DOWN": 22,
    "KEYTYPE_ILLUMINATION_TOGGLE": 23,
}



import logging
import os
import time
from typing import Dict
from typing import List
from typing import Union

from macuitest.lib.applescript_lib.applescript_wrapper import AppleScriptError
from macuitest.lib.applescript_lib.applescript_wrapper import as_wrapper
from macuitest.lib.core import wait_condition
from macuitest.lib.elements.applescript_element import ASElement
from macuitest.lib.operating_system.env import env
from macuitest.lib.operating_system.macos import macos
from macuitest.lib.operating_system.plist_helper import PlistHelper


class Application:
    def __init__(self, app_name: str, location: str = env.applications):
        self.name: str = app_name
        self.application_property_list: str = f"{location}/{self.name}.app/Contents/Info.plist"
        self.contents_reader: PlistHelper = PlistHelper(self.application_property_list)
        self.window = ASElement("window 1", process=self.name)

    def close_windows(self):
        try:
            return as_wrapper.tell_app(self.name, "close every window")
        except AppleScriptError:
            pass

    def relaunch(self) -> None:
        self.quit()
        time.sleep(3)
        self.launch()

    def quit(self, pause: Union[int, float] = 2) -> None:
        time.sleep(pause)
        as_wrapper.tell_app(self.name, "quit", ignoring_responses=True)
        if not macos.service_manager.wait_process_disappeared(self.name):
            macos.service_manager.kill_process(self.name)
        if not self.did_quit:
            raise EnvironmentError(f'{self.name} has not quit.')

    @property
    def did_quit(self) -> bool:
        return macos.service_manager.wait_process_disappeared(self.name)

    def launch(self) -> bool:
        macos.service_manager.launch_application_by_name(self.name)
        return self.window.wait_displayed(timeout=15)

    def activate(self):
        as_wrapper.tell_app(app=self.name, command="activate")
        if not self.frontmost:
            raise EnvironmentError(f'{self.name} has not activated')

    def set_window_position(self, position: str = "{1210, 670}"):
        if self.frontmost:
            try:
                return as_wrapper.tell_app_process(
                    f"set position of the front window to {position}", self.name
                )
            except AppleScriptError:
                pass

    def set_window_size(self, size: str = "{700, 400}"):
        if self.frontmost:
            try:
                return as_wrapper.tell_app_process(
                    f"set size of the front window to {size}", self.name
                )
            except AppleScriptError:
                pass

    @property
    def version(self) -> int:
        return self.contents_reader.read_property("CFBundleShortVersionString")

    @property
    def build(self) -> int:
        return self.contents_reader.read_property("CFBundleVersion")

    @property
    def is_running(self) -> bool:
        return macos.service_manager.wait_process_appeared(self.name, timeout=2)

    @property
    def did_launch(self) -> bool:
        return macos.service_manager.wait_process_appeared(self.name, timeout=20)

    def set_frontmost(self, value: bool = True):
        converted = {True: "true", False: "false"}.get(value)
        self._set_attribute("AXFrontmost", converted)

    def is_frontmost(self) -> bool:
        return wait_condition(lambda: self._read_attribute("AXFrontmost"), timeout=5)

    def set_hidden(self, value: bool = True):
        converted = {True: "true", False: "false"}.get(value)
        self._set_attribute("AXHidden", converted)

    def is_hidden(self) -> bool:
        return wait_condition(lambda: self._read_attribute("AXHidden"), timeout=3)

    def _set_attribute(self, attribute, value):
        return self.__execute(f'set value of attribute "{attribute}"', params=f"to {value}")

    def _read_attribute(self, attribute):
        try:
            return self.__execute(f'get value of attribute "{attribute}"')
        except AppleScriptError:
            return None

    def __execute(self, command, params=""):
        """Execute a command.
        :param str command: The name of the command to _execute as a string.
        :param str params: Command parameters.
        :return str: Execution output."""
        _command = f"{command} {params}" if params else f"{command}"
        return as_wrapper.tell_app_process(command=_command, app_process=self.name)

    hidden = property(is_hidden, set_hidden)
    frontmost = property(is_frontmost, set_frontmost)


class Finder(Application):
    def __init__(self):
        super().__init__("Finder")

    @property
    def window_path(self) -> str:
        return as_wrapper.tell_app(
            self.name, "return POSIX path of (target of first window as alias)"
        )

    def eject_mounted_disks(self) -> None:
        return as_wrapper.tell_app(self.name, "eject (every disk whose ejectable is true)")

    def move_file_to_trash(self, target: str) -> None:
        as_wrapper.tell_app(self.name, f'move POSIX file "{target}" to trash')

    def show_destination(self, destination: str) -> None:
        if not os.path.exists(destination):
            raise FileNotFoundError(destination)
        as_wrapper.tell_app(self.name, f'reveal POSIX file "{destination}"')
        self.window.wait_displayed()
        self.activate()

    def open_destination(self, destination: str) -> None:
        if not os.path.exists(destination):
            raise FileNotFoundError(destination)
        as_wrapper.tell_app(self.name, f'open POSIX file "{destination}"')


class Installer(Application):
    def __init__(self):
        super().__init__("Installer")

    def open_destination(self, destination: str) -> None:
        destination = destination.strip()
        if not os.path.exists(destination):
            raise FileNotFoundError(destination)
        macos.shell_executor.execute(f'open "{destination}" -a Installer')
        self.window.wait_displayed(timeout=10)
        self.activate()


class SystemPreferences(Application):
    anchors: Dict[str, str] = {
        "full_disk_access": "Privacy_AllFiles",
        "camera_access": "Privacy_Camera",
    }

    def __init__(self):
        super().__init__("System Preferences")

    def authorize(self):
        return as_wrapper.tell_app(self.name, "tell current pane to authorize")

    def show_anchor(self, anchor: str, pane_id="com.apple.preference.security"):
        return as_wrapper.tell_app(self.name, f'reveal anchor "{anchor}" of pane "{pane_id}"')

    def reveal_pane(self, pane_id="com.apple.preference.security"):
        return as_wrapper.tell_app(self.name, f'reveal pane "{pane_id}"')

    def get_pane_anchors(self, pane_id: str) -> List[str]:
        return as_wrapper.tell_app(self.name, f'return name of every anchor of pane "{pane_id}"')

    @property
    def current_pane_anchors(self) -> List[str]:
        try:
            return as_wrapper.tell_app(self.name, "return name of every anchor of current pane")
        except AppleScriptError:
            logging.warning("You must launch System Preferences first.")
            return list()

    @property
    def current_pane_id(self):
        return as_wrapper.tell_app(self.name, "return id of current pane")

    @property
    def pane_ids(self) -> List[str]:
        return as_wrapper.tell_app(self.name, "return id of every pane")


finder = Finder()
installer = Installer()
system_preferences_app = SystemPreferences()



from pathlib import Path
from typing import Optional
from typing import Tuple
from typing import Union

import AppKit
import numpy
import Quartz
from Foundation import NSURL
from Quartz import CGDisplayBounds
from Quartz import CGMainDisplayID
from Quartz import CoreGraphics

from macuitest.config.constants import Region
from macuitest.config.constants import ScreenSize


class Monitor:
    def __init__(self):
        self.__is_retina: Optional[bool] = None
        self.__screen_size: Optional[ScreenSize] = None

    def make_snapshot(self, region: Optional[Region] = None) -> numpy.ndarray:
        h, w = (
            (region.y2 - region.y1, region.x2 - region.x1)
            if region
            else (self.size.height, self.size.width)
        )
        pixel_data = self.get_pixel_data(region=region)
        _image = numpy.frombuffer(pixel_data[0], dtype=numpy.uint8)
        return _image.reshape((h, pixel_data[1], 4))[:, :w, :]

    @property
    def bytes(self):
        return bytes(numpy.frombuffer(self.get_pixel_data()[0], numpy.uint8))

    @property
    def is_retina(self) -> bool:
        if self.__is_retina is None:
            self.__is_retina = AppKit.NSScreen.mainScreen().backingScaleFactor() > 1.0
        return self.__is_retina

    @property
    def size(self) -> ScreenSize:
        if self.__screen_size is None:
            size = CGDisplayBounds(CGMainDisplayID()).size
            self.__screen_size = ScreenSize(int(size.width), int(size.height))
        return self.__screen_size

    @staticmethod
    def get_pixel_data(region: Optional[Region] = None):
        region = (
            CoreGraphics.CGRectInfinite
            if region is None
            else CoreGraphics.CGRectMake(
                region.x1, region.y1, region.x2 - region.x1, region.y2 - region.y1
            )
        )
        image = CoreGraphics.CGWindowListCreateImage(
            region,
            CoreGraphics.kCGWindowListOptionOnScreenOnly,
            CoreGraphics.kCGNullWindowID,
            CoreGraphics.kCGWindowImageDefault,
        )
        pixel_data = CoreGraphics.CGDataProviderCopyData(CoreGraphics.CGImageGetDataProvider(image))
        bytes_per_row = CoreGraphics.CGImageGetBytesPerRow(image) // 4
        return pixel_data, bytes_per_row

    @staticmethod
    def save_screenshot(
        where: Union[str, Path], region: Optional[Tuple[int, int, int, int]] = None
    ) -> Union[str, Path]:
        """Take a screenshot and save it to `where.
        Note: Region is defined by (x, y) pair of top left point, and width, length params.
        """
        region = CoreGraphics.CGRectInfinite if region is None else CoreGraphics.CGRectMake(*region)
        image = CoreGraphics.CGWindowListCreateImage(
            region,
            CoreGraphics.kCGWindowListOptionOnScreenOnly,
            CoreGraphics.kCGNullWindowID,
            CoreGraphics.kCGWindowImageDefault,
        )
        destination = Quartz.CGImageDestinationCreateWithURL(
            NSURL.fileURLWithPath_(str(where)), "public.png", 1, None
        )
        Quartz.CGImageDestinationAddImage(destination, image, dict())
        Quartz.CGImageDestinationFinalize(destination)
        return where


monitor = Monitor()



import os
from typing import Optional
from typing import Tuple

from macuitest.lib.operating_system.shell_executor import ShellExecutor


class Env:
    """macOS environment variables container."""

    _home: Optional[str] = None
    _version: Optional[str] = None

    def __init__(self):
        self.executor = ShellExecutor()

    @property
    def version_str(self) -> str:
        return ".".join(str(i) for i in env.version).strip()

    @property
    def version_major_str(self) -> str:
        return ".".join(str(i) for i in self.version_major).strip()

    @property
    def version_major(self) -> Tuple[int, ...]:
        return self.version[:2]

    @property
    def version(self) -> Tuple[int, ...]:
        """Get macOS version."""
        return tuple((int(item) for item in self.__version.split(".")))

    @property
    def __version(self) -> str:
        if self._version is None:
            self._version = self.executor.get_output("sw_vers -productVersion")
        return self._version

    @property
    def applications(self) -> str:
        return "/Applications"

    @property
    def system_applications(self) -> str:
        return f"{self.system}{self.applications}"

    @property
    def temp_dir(self) -> str:
        return os.getenv("TMPDIR")

    @property
    def system(self) -> str:
        return "/System"

    @property
    def system_library(self) -> str:
        return "/Library"

    @property
    def system_app_support(self) -> str:
        return os.path.join(self.system_library, "Application Support")

    @property
    def system_caches(self) -> str:
        return os.path.join(self.system_library, "Caches")

    @property
    def system_diag_reports(self) -> str:
        return os.path.join(self.system_logs, "DiagnosticReports")

    @property
    def system_extensions(self) -> str:
        return os.path.join(self.system_library, "Extensions")

    @property
    def system_internet_plugins(self) -> str:
        return os.path.join(self.system_library, "Internet Plug-Ins")

    @property
    def system_launch_daemons(self) -> str:
        return os.path.join(self.system_library, "LaunchDaemons")

    @property
    def system_logs(self) -> str:
        return os.path.join(self.system_library, "Logs")

    @property
    def system_preferences(self) -> str:
        return os.path.join(self.system_library, "Preferences")

    @property
    def system_preference_panes(self) -> str:
        return os.path.join(self.system_library, "PreferencePanes")

    @property
    def system_privileged_helper_tools(self) -> str:
        return os.path.join(self.system_library, "PrivilegedHelperTools")

    @property
    def system_security(self) -> str:
        return os.path.join(self.system_library, "Security")

    @property
    def system_widgets(self) -> str:
        return os.path.join(self.system_library, "Widgets")

    @property
    def user(self) -> str:
        return os.environ.get("USER")

    @property
    def home(self) -> str:
        return self.__home

    @property
    def trash(self) -> str:
        return os.path.join(self.__home, ".Trash")

    @property
    def desktop(self) -> str:
        return os.path.join(self.__home, "Desktop")

    @property
    def documents(self) -> str:
        return os.path.join(self.__home, "Documents")

    @property
    def downloads(self) -> str:
        return os.path.join(self.__home, "Downloads")

    @property
    def movies(self) -> str:
        return os.path.join(self.__home, "Movies")

    @property
    def music(self) -> str:
        return os.path.join(self.__home, "Music")

    @property
    def pictures(self) -> str:
        return os.path.join(self.__home, "Pictures")

    @property
    def user_caches(self) -> str:
        return os.path.join(self.user_lib, "Caches")

    @property
    def user_diag_reports(self) -> str:
        return os.path.join(self.user_lib, "Logs", "DiagnosticReports")

    @property
    def user_logs(self) -> str:
        return os.path.join(self.user_lib, "Logs")

    @property
    def user_launch_agents(self) -> str:
        return os.path.join(self.user_lib, "LaunchAgents")

    @property
    def user_app_support(self) -> str:
        return os.path.join(self.user_lib, "Application Support")

    @property
    def user_app_script(self) -> str:
        return os.path.join(self.user_lib, "Application Scripts")

    @property
    def user_saved_app_state(self) -> str:
        return os.path.join(self.user_lib, "Saved Application State")

    @property
    def user_cookies(self) -> str:
        return os.path.join(self.user_lib, "Cookies")

    @property
    def user_containers(self) -> str:
        return os.path.join(self.user_lib, "Containers")

    @property
    def user_group_containers(self) -> str:
        return os.path.join(self.user_lib, "Group Containers")

    @property
    def user_preferences(self) -> str:
        return os.path.join(self.user_lib, "Preferences")

    @property
    def user_lib(self) -> str:
        return os.path.join(self.__home, "Library")

    @property
    def __home(self) -> str:
        """Get path to the user HOME folder."""
        if self._home is None:
            self._home = os.getenv("HOME")
        return self._home


env = Env()



from pathlib import Path
from typing import Optional
from typing import Union

import cv2

from macuitest.config.constants import Point
from macuitest.config.constants import Region
from macuitest.lib.core import wait_condition
from macuitest.lib.elements.controllers.mouse import MouseConfig
from macuitest.lib.elements.controllers.mouse import mouse
from macuitest.lib.elements.ui.monitor import monitor


class UIElementNotFoundOnScreen(Exception):
    """Thrown when a pattern is not found on the screen."""


class UIElement:
    """Visible user interface element. Based on automated pattern lookup algorithm (OpenCV)."""

    def __init__(self, screenshot_path: Union[str, Path], similarity: float = 0.925):
        self.path = screenshot_path.strip() if isinstance(screenshot_path, str) else screenshot_path
        self.similarity: float = similarity
        self.image, self.width, self.height = None, None, None
        self.__matches: Optional = list()
        self.__load_image()

    def __repr__(self):
        return f'<UIElement "{self.path}", similarity={self.similarity}>'

    def paste(
        self, x_off: int = 0, y_off: int = 0, phrase: str = "", region: Optional[Region] = None
    ):
        center = self.get_center(region)
        mouse.paste(center.x + x_off, center.y + y_off, phrase=phrase)

    def double_click(self, x_off: int = 0, y_off: int = 0, region: Optional[Region] = None):
        center = self.get_center(region)
        mouse.double_click(center.x + x_off, center.y + y_off)

    def right_click(
        self,
        x_off: int = 0,
        y_off: int = 0,
        hold: float = MouseConfig.hold,
        pause: float = MouseConfig.pause,
        region: Optional[Region] = None,
    ):
        center = self.get_center(region)
        mouse.right_click(center.x + x_off, center.y + y_off, hold=hold, pause=pause)

    def click_mouse(
        self,
        x_off: int = 0,
        y_off: int = 0,
        hold: float = MouseConfig.hold,
        pause: float = MouseConfig.pause,
        region: Optional[Region] = None,
    ):
        center = self.get_center(region)
        mouse.click(center.x + x_off, center.y + y_off, hold=hold, pause=pause)

    def hover_mouse(
        self,
        x_off: int = 0,
        y_off: int = 0,
        duration: float = MouseConfig.move,
        region: Optional[Region] = None,
    ):
        center = self.get_center(region)
        mouse.hover(center.x + x_off, center.y + y_off, duration=duration)

    @property
    def is_visible(self) -> bool:
        """Check whether the pattern is visible on the screen."""
        return False or self.wait_displayed() is not None

    def get_center(self, region: Optional[Region] = None):
        match = self.wait_displayed(region=region)
        if not match:
            raise UIElementNotFoundOnScreen(self.path)
        return Point(int(match.x + self.width / 2), int(match.y + self.height / 2))

    def wait_displayed(
        self, timeout: int = 5, region: Optional[Region] = None
    ) -> Union[None, Point]:
        return wait_condition(lambda: self.detect_on_screen(region), timeout=timeout)

    def wait_vanish(self, timeout: int = 15, region: Optional[Region] = None) -> bool:
        return wait_condition(lambda: self.detect_on_screen(region) is None, timeout=timeout)

    def detect_on_screen(self, region: Optional[Region] = None):
        """Locate pattern on the screen and return its center.
        OpenCV (Open Source Computer Vision Library) is an open source computer vision
        and machine learning software library. It's built to provide a common infrastructure
        for computer vision applications and to accelerate the use of machine perception
        in the commercial products."""
        region = region or Region(0, 0, monitor.size.width, monitor.size.height)
        _, similarity, _, position = cv2.minMaxLoc(
            cv2.matchTemplate(
                cv2.cvtColor(monitor.make_snapshot(region), cv2.COLOR_BGR2GRAY),
                self.image,
                cv2.TM_CCOEFF_NORMED,
            )
        )
        # In case of AssertionError in cv2.error:
        # `_img.size().height <= _templ.size().height && _img.size().width <= _templ.size().width`
        # You need to check that the `region` size is larger than `pattern` size.
        if round(similarity, 3) >= self.similarity:
            denominator = 2 if monitor.is_retina else 1
            return Point(
                (position[0] + region.x1) // denominator, (position[1] + region.y1) // denominator
            )

    def __load_image(self) -> None:
        """Load the image from the disk."""
        if not Path(self.path).exists():
            raise FileNotFoundError(f"Cannot find request screenshot: {self.path}")
        image = cv2.imread(self.path, cv2.IMREAD_GRAYSCALE)
        if image is None:
            raise IOError(f"Cannot not load screenshot: {self.path}")
        height, width = image.shape
        if monitor.is_retina:
            height, width = height / 2, width / 2
        self.image, self.width, self.height = image, width, height



import threading
from typing import Dict
from typing import List
from typing import Optional

import biplist

from macuitest.lib import core


class MemoryManager:
    """Interface to memory (RAM) information and manipulations on it."""

    def __init__(self, executor):
        self._memory_slots: Optional[List[Dict[str, str]]] = None
        self.executor = executor

    def load_memory(self, percent_free: int, timeout: int = 15) -> None:
        """Simulate memory pressure on the system in background."""
        threading.Thread(
            target=self.executor.execute, args=(f"memory_pressure -p {percent_free}", timeout)
        ).start()

    @property
    def total_memory_bytes(self) -> int:
        return core.convert_file_size_to_bytes(self.total_memory_str)

    @property
    def total_memory_str(self) -> str:
        return f"{self.total_memory} GB"

    @property
    def total_memory(self) -> int:
        return sum((int(bank.get("dimm_size").split()[0]) for bank in self.memory_slots))

    @property
    def memory_banks(self) -> int:
        return len(self.memory_slots)

    @property
    def memory_speed(self) -> str:
        return self.memory_slots[0].get("dimm_speed")

    @property
    def memory_type(self) -> str:
        return self.memory_slots[0].get("dimm_type")

    @property
    def memory_slots(self) -> List[Dict[str, str]]:
        if self._memory_slots is not None:
            return self._memory_slots
        _info = bytes(
            self.executor.get_output("system_profiler -xml SPMemoryDataType"), encoding="utf-8"
        )
        _slots = biplist.readPlistFromString(_info)[0].get("_items")[0].get("_items")
        self._memory_slots = _slots
        return _slots



import logging
import os
import subprocess
from typing import Optional
from typing import Union


class ShellExecutor:
    execution_timeout_code: int = -1001
    pipe: int = -1
    stdout: int = -2
    devnull: int = -3

    def __init__(self):
        self.__password: Optional[str] = None

    def sudo_get_output(self, cmd: str, timeout: Optional[int] = None) -> Optional[str]:
        """Execute shell instruction with elevated privileges and return execution output."""
        response = self.__execute(
            cmd=f'echo "{self.get_admin_password()}" | sudo -S -p "" {cmd}', timeout=timeout
        )
        if isinstance(response, subprocess.CompletedProcess):
            return response.stdout.decode(encoding="utf-8").strip()

    def get_output(self, cmd: str, timeout: Optional[int] = None) -> Optional[str]:
        """Execute shell instruction and return execution output."""
        response = self.__execute(cmd=cmd, timeout=timeout)
        if isinstance(response, subprocess.CompletedProcess):
            return response.stdout.decode(encoding="utf-8").strip()

    def sudo(
        self, cmd: str, timeout: Optional[int] = None
    ) -> Union[subprocess.CompletedProcess, int]:
        """Execute shell instruction with elevated privileges and return execution code."""
        return self.__execute(
            cmd=f'echo "{self.get_admin_password()}" | sudo -S -p "" {cmd}', timeout=timeout
        )

    def execute(
        self, cmd: str, timeout: Optional[int] = None
    ) -> Union[subprocess.CompletedProcess, int]:
        """Execute a shell instruction and return a `CompletedProcess` object."""
        return self.__execute(cmd=cmd, timeout=timeout)

    def get_admin_password(self, auto_keychain: str = "com.macuitest.automation"):
        if self.__password is None:
            subprocess.call("sudo -k", shell=True)
            _password = subprocess.getoutput(
                f"security find-generic-password -a {auto_keychain} -w 2>/dev/null"
            )
            if (
                subprocess.call(f'echo "{_password}" |sudo -S -p "" -v --', shell=True, stderr=-3)
                == 0
            ):
                self.__password = _password
            else:  # Fallback. Reading environment variables.
                _password = os.environ.get("MACUITEST_PASSWORD")
                if (
                    subprocess.call(
                        f'echo "{_password}" |sudo -S -p "" -v --', shell=True, stderr=-3
                    )
                    == 0
                ):
                    self.__password = _password
                else:
                    raise SystemExit(
                        f"\nCANNOT USE SPECIFIED ADMIN PASSWORD:\n"
                        f"a) Add your actual password to environment variables\n"
                        f"b) Create a login keychain item with account name "
                        f'"{auto_keychain}" in Keychain Access.app\n'
                        f"https://support.apple.com/en-gb/guide/keychain-access/kyca1120/mac\n"
                    )
        return self.__password

    def __execute(
        self, cmd: str, timeout: Optional[int] = None
    ) -> Union[subprocess.CompletedProcess, int]:
        """Send instruction to shell."""
        try:
            response = subprocess.run(
                cmd, shell=True, stderr=self.pipe, stdout=self.pipe, timeout=timeout
            )
            if response.returncode:
                warning_message = (
                    f'Command "{cmd}" failed with exit code {response.returncode}\n\t'
                    f'{response.stderr.decode(encoding="utf-8").strip()}'
                )
                logging.debug(warning_message)
            return response
        except subprocess.TimeoutExpired as e:
            logging.error(e)
            return self.execution_timeout_code



import time
from typing import ClassVar
from typing import Tuple

from macuitest.config.constants import DNSMapping


class Hosts:
    hosts_file: ClassVar[str] = "/etc/hosts"

    def __init__(self, executor, file_manager):
        self.executor = executor
        self.file_manager = file_manager

    def change_mapping(self, mapping: DNSMapping):
        """Configure local domain resolution for a certain resource."""
        self.file_manager.change_file_permissions(
            path=self.hosts_file, permissions="777", sudo=True
        )
        with open(self.hosts_file) as hosts:
            hosts_config = hosts.read().rstrip().splitlines()
        hosts_config.append(mapping)
        updated = "\n".join(hosts_config)
        with open(self.hosts_file, "w") as hosts:
            hosts.write(f"{updated}\n")
        self.file_manager.change_file_permissions(
            path=self.hosts_file, permissions="755", sudo=True
        )

    def reset(self, known_mappings: Tuple[DNSMapping, ...], disable_pause: bool = False):
        """Reset /etc/hosts by removing certain/all known mappings."""
        self.file_manager.change_file_permissions(
            path=self.hosts_file, permissions="777", sudo=True
        )
        with open(self.hosts_file, "r") as hosts:
            hosts_config = hosts.read().rstrip().splitlines()
        updated = "\n".join((line for line in hosts_config if line not in known_mappings))
        with open(self.hosts_file, "w") as hosts:
            hosts.write(f"{updated}\n")
        self.file_manager.change_file_permissions(
            path=self.hosts_file, permissions="755", sudo=True
        )
        self.flush_dns_cache(disable_pause=disable_pause)

    def flush_dns_cache(self, disable_pause: bool = False):
        """Flush DNS cache to remove stored website location details."""
        time.sleep(0 if disable_pause else 2)
        self.executor.execute("dscacheutil -flushcache")
        time.sleep(0 if disable_pause else 1)
        self.executor.sudo("killall -HUP mDNSResponder")
        time.sleep(0 if disable_pause else 5)  # For macOS to catch up.



"""Some basic functions to be used throughout the project."""
import functools
import time
from typing import Any
from typing import Callable
from typing import Tuple
from typing import Union


class WaitConditionException(Exception):
    """A `placeholder` exception for `wait_condition`."""


def wait_condition(
    predicate: Callable,
    timeout: Union[int, float] = 10,
    exceptions: tuple = (WaitConditionException,),
    *args,
    **kwargs,
) -> Any:
    """Execute `predicate` and return the result."""
    end = time.time() + timeout
    while time.time() < end:
        time.sleep(0.005)
        try:
            result = predicate(*args, **kwargs)
            if result:
                return result
        except exceptions:
            continue
    return False


def _parametrized(decorator):
    @functools.wraps(decorator)
    def wrapper(*args, **kwargs):
        def result(func):
            return decorator(func, *args, **kwargs)

        return result

    return wrapper


@_parametrized
def slow_down(func, seconds: float):
    """Slow down `func` by adding timeouts before and after its execution."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        time.sleep(seconds)
        result = func(*args, **kwargs)
        time.sleep(seconds)
        return result

    return wrapper


def is_close(
    actual: Union[int, float], expected: Union[int, float], threshold: Union[int, float]
) -> bool:
    """Compare the difference between two input numbers with specified threshold."""
    return abs(round(actual, 2) - round(expected, 2)) <= threshold


def convert_version_from_tuple_to_str(string: Tuple[int, ...]) -> str:
    """Convert program version from tuple to string."""
    return ".".join(str(i) for i in string)


def convert_version_from_str_to_tuple(string: str) -> Tuple[int, ...]:
    """Convert program version from string to tuple."""
    return tuple([int(item) for item in string.split(".")])


def convert_file_size_to_bytes(string: str) -> int:
    """Convert human readable file size to integer."""
    _size, _unit = string.strip().lower().replace("zero", "0").replace(",", "").split()
    dots = len([i for i in _size if i == "."])
    return int(
        float(_size.replace(".", "", dots - 1))
        * 1000 ** ("bytes", "kb", "mb", "gb", "tb").index(_unit)
    )



import glob
import logging
import os
import time
from typing import ClassVar
from typing import List

from macuitest.config.constants import MINUTE
from macuitest.lib.applescript_lib.applescript_wrapper import AppleScriptError
from macuitest.lib.applescript_lib.applescript_wrapper import as_wrapper
from macuitest.lib.apps.application import Application
from macuitest.lib.core import wait_condition
from macuitest.lib.elements.native.calls import AXErrorInvalidUIElement
from macuitest.lib.elements.native.native_ui_element import NativeUIElement
from macuitest.lib.operating_system.env import env

EXCEPTIONS = (AttributeError, ValueError, AXErrorInvalidUIElement)


class Safari(Application):
    def __init__(self, app_name="Safari"):
        super().__init__(app_name)
        self.bundle_id: ClassVar[str] = "com.apple.Safari"

    def download_file(self, url: str, file_name: str = "*pattern*", timeout: int = MINUTE):
        _before_ = self.downloads
        self.request_webpage(url)
        self.confirm_download()
        wait_condition(lambda: self.downloads > _before_, timeout=MINUTE)
        wait_condition(
            lambda: glob.glob(os.path.join(env.downloads, "*.download")) == [], timeout=timeout
        )
        return wait_condition(lambda: glob.glob(os.path.join(env.downloads, file_name)), timeout=30)

    def request_webpage(self, url: str):
        as_wrapper.tell_app(self.name, f'make new document with properties {{URL:"{url}"}}')
        wait_condition(lambda: self.native_window is not None)

    def close_tabs(self):
        if not self.is_running:
            return
        try:
            as_wrapper.tell_app(self.name, "close tabs of every window", ignoring_responses=True)
            wait_condition(
                lambda: as_wrapper.tell_app(self.name, "return the number of tabs in windows") == 0
            )
        except AssertionError:
            as_wrapper.tell_app(self.name, "close every window", ignoring_responses=True)

    def did_webpage_load(self, expected_address: str, timeout: float = 15) -> bool:
        if not self.did_launch:
            raise EnvironmentError(f'{self.name} did not launch')
        wait_condition(lambda: self.native_window is not None, exceptions=EXCEPTIONS)
        wait_condition(
            lambda: self.execute_js_command("document.readyState") == "complete", timeout=timeout
        )
        if not self.did_reload_button_appear:
            logging.warning("Reload button has not shown")
        return wait_condition(
            lambda: expected_address in self.document_url, timeout=3, exceptions=EXCEPTIONS
        )

    def execute_js_command(self, command: str):
        command = command.replace('"', '\\"')
        tell_what = f'tell front document to do JavaScript "{command}"'
        try:
            return as_wrapper.tell_app(self.name, tell_what)
        except AppleScriptError:
            logging.warning(f'Could not execute: "{tell_what}"')

    def search_web(self, query: str) -> None:
        as_wrapper.tell_app(self.name, f'tell front tab to search the web for "{query}"')

    def confirm_download(self) -> None:
        time.sleep(0.5)
        confirm_download_dialog = self.confirm_download_dialog
        if confirm_download_dialog:
            confirm_download_dialog.find_element(AXRole="AXButton", AXTitle="Allow").press()

    @property
    def document_cookies(self):
        return self.execute_js_command("document.cookie")

    @property
    def document_url(self) -> str:
        return str(as_wrapper.tell_app(self.name, "tell front document to return URL"))

    @property
    def document_name(self) -> str:
        return str(as_wrapper.tell_app(self.name, "tell front document to return name"))

    @property
    def document_html(self) -> str:
        return str(as_wrapper.tell_app(self.name, "tell front document to return source"))

    def __get_address_bar_value(self) -> str:
        return self.address_bar.get_ax_attribute("AXValue")

    def __set_address_bar_value(self, new_value: str) -> None:
        address_bar = self.address_bar
        address_bar.set_ax_attribute("AXFocused", True)
        time.sleep(0.5)
        address_bar.set_ax_attribute("AXValue", new_value)

    address_bar_value = property(__get_address_bar_value, __set_address_bar_value)

    @property
    def did_reload_button_appear(self) -> bool:
        return wait_condition(
            lambda: self.stop_reload_button.get_ax_attribute("AXTitle") == "Reload this page",
            exceptions=EXCEPTIONS,
        )

    @property
    def stop_reload_button(self) -> NativeUIElement:
        return self.native_window.find_element(AXRole="AXToolbar").find_element(
            AXIdentifier="StopReloadButton", recursive=True
        )

    @property
    def address_bar(self) -> NativeUIElement:
        return self.native_window.find_element(
            AXIdentifier="WEB_BROWSER_ADDRESS_AND_SEARCH_FIELD", recursive=True
        )

    @property
    def tabs(self) -> List[NativeUIElement]:
        try:
            result = self.native_window.find_element(AXIdentifier="TabBar").get_ax_attribute(
                "AXChildren"
            )
        except AttributeError:
            result = []
        return result

    @property
    def confirm_download_dialog(self) -> NativeUIElement:
        return self.native_window.find_element(AXIdentifier="AXSafariModalDialog", recursive=True)

    @property
    def native_window(self) -> NativeUIElement:
        return NativeUIElement.from_bundle_id(self.bundle_id).find_element(
            AXRole="AXWindow", AXMain=True
        )

    @property
    def downloads(self) -> List[str]:
        return [f for f in os.listdir(env.downloads) if not f.startswith(".")]


safari = Safari()



import math
import time
from dataclasses import dataclass
from datetime import datetime
from types import MappingProxyType
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from macuitest.config.constants import CheckboxState
from macuitest.config.constants import DisclosureTriangleState
from macuitest.config.constants import Frame
from macuitest.config.constants import Point
from macuitest.config.constants import Region
from macuitest.lib import core
from macuitest.lib.applescript_lib.applescript_wrapper import AppleScriptError
from macuitest.lib.applescript_lib.applescript_wrapper import as_wrapper
from macuitest.lib.core import is_close
from macuitest.lib.core import wait_condition
from macuitest.lib.elements.controllers.keyboard_controller import keyboard
from macuitest.lib.elements.controllers.mouse import MouseConfig
from macuitest.lib.elements.controllers.mouse import mouse
from macuitest.lib.elements.ui.monitor import monitor
from macuitest.lib.operating_system.color_meter import get_color
from macuitest.lib.operating_system.color_meter import get_most_common_color
from macuitest.lib.operating_system.env import env


class BaseUIElement:
    """Represent a UI element based on an AppleScript locator."""

    __slots__ = ("locator", "process", "_frame")

    def __init__(self, locator: str, process: str):
        self.locator = locator
        self.process = process

    def __repr__(self):
        return f'<{self.__class__.__name__} "{self.locator}", process="{self.process}">'

    def wait_on_position(self, position: Point) -> bool:
        for _ in range(50):
            f = self.frame
            if is_close(f.center.x, position.x, threshold=2) and is_close(
                f.center.y, position.y, threshold=2
            ):
                return True
            time.sleep(0.1)

    def snapshot(self, margin: int = 4) -> str:
        frame = self.frame
        element_region = (
            frame.x1 - margin,
            frame.y1 - margin,
            frame.width + margin * 2,
            frame.height + margin * 2,
        )
        return monitor.save_screenshot(
            region=element_region,
            where=f'{env.desktop}/scr_{datetime.now().strftime("%Y%m%d%H%M%S%f")}.png',
        )

    @property
    def region(self):
        return self.get_region()

    def get_region(self, margin: int = 0) -> Region:
        f = self.frame
        return Region(f.x1 - margin, f.y1 - margin, f.x2 + margin, f.y2 + margin)

    def most_common_color(self, ignore_colors: Optional[Tuple[str, ...]] = None):
        f = self.frame
        return get_most_common_color(f.x1, f.x2, f.y1, f.y2, ignore_colors)

    def color(self, x_off: int = 0, y_off: int = 0) -> str:
        c = self.frame.center
        return get_color(Point(c.x + x_off, c.y + y_off)).replace("gray", "grey")

    def scroll(self, x_off: int = 0, y_off: int = 0, clicks: int = 1):
        c = self.frame.center
        mouse.scroll(c.x + x_off, c.y + y_off, scrolls=clicks)

    def doubleclick_mouse(self, x_off: int = 0, y_off: int = 0, duration: float = MouseConfig.move):
        c = self.frame.center
        mouse.double_click(c.x + x_off, c.y + y_off, duration=duration)

    def rightclick_mouse(
        self,
        x_off: int = 0,
        y_off: int = 0,
        hold: float = MouseConfig.hold,
        duration: float = MouseConfig.move,
        pause: float = MouseConfig.pause,
    ) -> None:
        c = self.frame.center
        mouse.right_click(c.x + x_off, c.y + y_off, hold, duration, pause)

    def click_mouse(
        self,
        x_off: int = 0,
        y_off: int = 0,
        hold: float = MouseConfig.hold,
        duration: float = MouseConfig.move,
        pause: float = MouseConfig.pause,
    ) -> None:
        c = self.frame.center
        mouse.click(c.x + x_off, c.y + y_off, hold, duration, pause)

    def hover_mouse(
        self, x_off: int = 0, y_off: int = 0, duration: float = MouseConfig.move
    ) -> None:
        c = self.frame.center
        mouse.hover(c.x + x_off, c.y + y_off, duration=duration)

    @property
    def frame(self) -> Frame:
        self.__assert_visible()
        _frame = (*self.get_attribute_value("AXPosition"), *self.get_attribute_value("AXSize"))
        x1, y1, width, height = (math.floor(x) for x in _frame)
        x2, y2 = x1 + width, y1 + height
        center = Point(int((x1 + width / 2)), int((y1 + height / 2)))
        return Frame(x1, y1, x2, y2, center, width, height)

    def click(self, pause: float = 0.4) -> bool:
        """Perform click action the element."""
        self.__assert_visible()
        time.sleep(pause)
        self._execute("click")
        time.sleep(0.25)
        return True

    def _select(self):
        """Click an element."""
        self.__assert_visible()
        return self._execute("select")

    def _set_focus(self, value):
        """Make element focused."""
        self.__assert_visible()
        return self._execute("set focused of", params=f'to "{value}"')

    def _show_context_menu(self):
        self.__assert_visible()
        return self._execute('perform action "AXShowMenu" of')

    def _set_value(self, value):
        """Set element value."""
        self.__assert_visible()
        return self._execute("set value of", params=f'to "{value}"')

    def _get_value(self):
        """Get element value."""
        self.__assert_visible()
        return self._execute("get value of")

    def _count_elements(self) -> int:
        """Count UI elements"""
        self.__assert_visible()
        return self._execute("count UI elements of")

    @property
    def _children(self):
        self.__assert_visible()
        return self.get_attribute_value("AXChildren")

    @property
    def _rows(self) -> int:
        self.__assert_visible()
        return self._execute("count rows of")

    @property
    def title(self) -> str:
        self.__assert_visible()
        return self.get_attribute_value("AXTitle").strip()

    @property
    def description(self) -> str:
        self.__assert_visible()
        return self.get_attribute_value("AXDescription").strip()

    @property
    def value(self) -> str:
        self.__assert_visible()
        return self.get_attribute_value("AXValue")

    @property
    def help(self) -> str:
        self.__assert_visible()
        return self.get_attribute_value("AXHelp")

    @property
    def _placeholder(self) -> str:
        self.__assert_visible()
        return self.get_attribute_value("AXPlaceholderValue")

    def _is_enabled(self) -> bool:
        """Check whether element enabled"""
        self.__assert_visible()
        return self.get_attribute_value("AXEnabled")

    @property
    def did_vanish(self) -> bool:
        return self.wait_vanish()

    def wait_vanish(self, timeout: [int, float] = 5) -> bool:
        self.wait_displayed(timeout=0.3)
        return wait_condition(lambda: self.is_exists() is False, timeout=timeout)

    @property
    def is_visible(self) -> bool:
        return self.wait_displayed()

    def wait_displayed(self, timeout: Union[int, float] = 5):
        return wait_condition(self.is_exists, timeout=timeout)

    def perform_action(self, action):
        self._execute(f'perform action "{action}" of')

    @property
    def actions(self) -> List[str]:
        try:
            return self._execute("get name of every action of")
        except AppleScriptError:
            return []

    def _set_attribute(self, attribute, value) -> Any:
        return self._execute(f'set value of attribute "{attribute}" of', params=f"to {value}")

    def get_attribute_value(self, attribute) -> Any:
        try:
            return self._execute(f'get value of attribute "{attribute}" of')
        except AppleScriptError:
            return None

    @property
    def attributes(self) -> List[str]:
        try:
            return self._execute("get name of every attribute of")
        except AppleScriptError:
            return []

    def is_exists(self) -> bool:
        try:
            return self._execute("return exists")
        except AppleScriptError as e:
            if e.number == -10000:
                pass
            else:
                raise

    def __assert_visible(self):
        if not self.is_visible:
            raise LookupError(self)

    def _execute(self, command: str, params: str = ""):
        """Execute a command.
        :param str command: The name of the command to _execute as a string.
        :param str params: Command parameters.
        :return str: Execution output."""
        _command = f"{command} {self.locator} {params}" if params else f"{command} {self.locator}"
        return as_wrapper.tell_app_process(command=_command, app_process=self.process)


class Button(BaseUIElement):
    @property
    def is_enabled(self):
        return wait_condition(self._is_enabled)

    def wait_enabled(self, timeout: int) -> bool:
        return wait_condition(self._is_enabled, timeout=timeout)

    @property
    def is_disabled(self):
        return wait_condition(lambda: self._is_enabled() is False)

    @property
    def value(self):
        return self._get_value()


class BusyIndicator(BaseUIElement):
    pass


class TextElement(BaseUIElement):
    def __repr__(self):
        try:
            return f"'{self.text}'"
        except (AppleScriptError, LookupError):
            return f"{self.__class__.__name__}(locator='{self.locator}', process='{self.process}')"

    @property
    def to_bytes(self) -> int:
        return core.convert_file_size_to_bytes(self.text)

    @property
    def characters_number(self):
        return self.get_attribute_value("AXNumberOfCharacters")

    @property
    def visible_characters_number(self):
        return list(self.get_attribute_value("AXVisibleCharacterRange"))[1]

    def set_text(self, value: str):
        self.wait_displayed()
        self._set_focus(True)
        self._set_value(value)

    def get_text(self) -> str:
        self.wait_displayed(timeout=1)
        return str(self._get_value())

    @property
    def is_fully_visible(self):
        self.wait_displayed(timeout=1)
        return self.characters_number == self.visible_characters_number

    def __eq__(self, other):
        return wait_condition(lambda: self.text == other, timeout=1)

    text = property(get_text, set_text)


class StaticText(TextElement):
    pass


class TextField(TextElement):
    def fill(self, with_text: str, pause: float = 0.01):
        for _ in range(2):
            self.text = ""
            self.focus()
            time.sleep(0.5)
            keyboard.write(with_text, pause=pause)
            if wait_condition(lambda: len(self.text) == len(with_text), timeout=0.5):
                break
        if not len(self.text) == len(with_text):
            raise KeyboardInterrupt(f'"{with_text}" expected but "{self.text}" received.')

    def focus(self, value="true"):
        self._set_focus(value)

    @property
    def is_secure(self) -> bool:
        return self.get_attribute_value("AXRoleDescription") == "secure text field"

    @property
    def placeholder(self):
        return self._placeholder

    def __get_value(self) -> str:
        return self._get_value()

    def __set_value(self, value):
        self._set_value(str(value))

    def __get_focused(self):
        return self.get_attribute_value("AXFocused")

    def __set_focused(self, value: bool):
        self._set_focus(value)

    is_focused = property(__get_focused, __set_focused)
    text = property(__get_value, __set_value)


class TextArea(TextField):
    pass


class ComboBox(BaseUIElement):
    def value(self) -> str:
        return self._get_value()

    def set_value(self, value):
        self.wait_displayed()
        time.sleep(0.5)
        self._set_value(value)

    text = property(value, set_value)


class Checkbox(BaseUIElement):
    @property
    def is_enabled(self):
        return wait_condition(self._is_enabled, timeout=3)

    @property
    def is_disabled(self):
        return wait_condition(lambda: self._is_enabled() is False)

    def wait_state(self, state):
        return wait_condition(lambda: self.get_state() == state)

    def set_state(self, state):
        target_state = CheckboxState.states.get(state)
        for _ in range(2):
            if self.get_state() != target_state:
                self.click()
        if self.state != target_state:
            raise EnvironmentError(f'Cannot set checkbox to state: "{state}"')

    def get_state(self):
        return self._get_value()

    state = property(get_state, set_state)


class DisclosureTriangle(BaseUIElement):
    def get_state(self):
        return self._get_value()

    def set_state(self, state: DisclosureTriangleState):
        attempts_left = 2
        while self.state != state and attempts_left:
            self.click()
            attempts_left -= 1
        if self.state != state:
            raise EnvironmentError(f'Cannot set disclosure triangle to state: "{state}"')

    state = property(get_state, set_state)


class MenuItem(BaseUIElement):
    @property
    def is_enabled(self):
        return wait_condition(self._is_enabled)


class MenuBarItem(BaseUIElement):
    pass


class Menu(BaseUIElement):
    pass


class Group(BaseUIElement):
    @property
    def elements_number(self):
        return self._count_elements()


class Row(BaseUIElement):
    def select(self, pause: float = 0.3):
        self._select()
        time.sleep(pause)

    @property
    def is_selected(self) -> bool:
        return self.get_attribute_value("AXSelected")


class ScrollArea(BaseUIElement):
    pass


class Table(BaseUIElement):
    @property
    def rows_number(self) -> int:
        return self._rows


class Image(BaseUIElement):
    @property
    def label(self):
        return self.description


class _List(BaseUIElement):
    @property
    def elements_number(self):
        return self._count_elements()


class Outline(BaseUIElement):
    @property
    def elements_number(self):
        return self._count_elements()


class Popover(BaseUIElement):
    pass


class Sheet(BaseUIElement):
    pass


class Window(BaseUIElement):
    def __init__(self, locator, process):
        super().__init__(locator, process)

    def set_position(self, point: Tuple[int, int]):
        self._execute("set position of", params="to {%s, %s}" % (point[0], point[1]))

    def get_position(self) -> Tuple[int, int]:
        return tuple(self._execute("get position of"))

    def set_size(self, size: Tuple[int, int]):
        self._execute("set size of", params="to {%s, %s}" % (size[0], size[1]))

    def get_size(self) -> Tuple[int, int]:
        return tuple(self._execute("get size of"))

    def set_minimized(self, value):
        converted = {True: "true", False: "false"}.get(value)
        self._set_attribute("AXMinimized", converted)

    def is_minimized(self):
        return self.get_attribute_value("AXMinimized")

    def set_full_screen(self, value):
        converted = {True: "true", False: "false"}.get(value)
        self._set_attribute("AXFrontmost", converted)

    def is_full_screen(self):
        return self.get_attribute_value("AXFullScreen")

    @property
    def is_keyboard_focused(self) -> bool:
        return self.get_attribute_value("AXFocused")

    position = property(get_position, set_position)
    size = property(get_size, set_size)
    minimized = property(is_minimized, set_minimized)
    full_screen = property(is_full_screen, set_full_screen)


class WebView(BaseUIElement):
    @property
    def url(self):
        wait_condition(lambda: self.get_attribute_value("AXURL") is not None, timeout=60)
        wait_condition(lambda: self.get_attribute_value("AXURL").startswith("http"), timeout=30)
        return wait_condition(lambda: self.get_attribute_value("AXURL"), timeout=30)


@dataclass(frozen=True)
class Elements:
    all: MappingProxyType = MappingProxyType(
        {
            "UI element": BaseUIElement,
            "button": Button,
            "pop up button": Button,
            "radio button": Button,
            "busy indicator": BusyIndicator,
            "static text": StaticText,
            "text area": TextArea,
            "text field": TextField,
            "checkbox": Checkbox,
            "combo box": ComboBox,
            "menu item": MenuItem,
            "menu button": MenuItem,
            "menu bar item": MenuBarItem,
            "menu": Menu,
            "group": Group,
            "row": Row,
            "scroll area": ScrollArea,
            "table": Table,
            "image": Image,
            "list": _List,
            "outline": Outline,
            "pop over": Popover,
            "popover": Popover,
            "sheet": Sheet,
            "window": Window,
        }
    )


class ASElement:
    """AppleScript element factory."""

    def __new__(
        cls, locator: str, process: str
    ) -> Union[
        BaseUIElement,
        Button,
        BusyIndicator,
        TextArea,
        TextField,
        Checkbox,
        ComboBox,
        MenuItem,
        Group,
        Row,
        ScrollArea,
        Table,
        Image,
        List,
        Outline,
        Popover,
        Sheet,
        Window,
    ]:
        loc = locator.split(" of")[0]
        loc = (
            "".join((i for i in loc if not i.isdigit())) if loc[-1].isdigit() else loc.split('"')[0]
        ).strip()
        return Elements.all.get(loc, BaseUIElement)(locator, process)



import json
import re
import time
from typing import ClassVar
from typing import Tuple
from urllib import error
from urllib import request


class Networking:
    ports_list: ClassVar[Tuple[str, ...]] = ("en0", "en1")  # Ethernet port and Wifi port.

    def __init__(self, executor):
        self.executor = executor

    def send_curl_request(self, request_):
        self.executor.execute(f"curl {request_}")

    @property
    def mac_address(self) -> str:
        _device = self.active_internet_device
        _mac_address_raw = self.executor.get_output(f"networksetup -getmacaddress {_device}")
        return (
            _mac_address_raw.replace("Ethernet Address: ", "")
            .replace(f" (Device: {_device})", "")
            .strip()
        )

    @property
    def active_internet_device(self) -> str:
        for device in self.network_service_order:
            _status = self.executor.get_output(f"ifconfig {device} |grep status").split(": ")[-1]
            if _status == "active":
                return device

    @property
    def network_service_order(self):
        _network_service_order = self.executor.get_output("networksetup listnetworkserviceorder")
        _devices = re.findall(r"Device: \w+-\w+|Device: \w+", _network_service_order)
        _devices = [d.split(": ")[-1] for d in _devices if "en" in d]
        return _devices

    def turn_internet_on(self):
        """Wait until the Internet connection is established."""
        for port in self.ports_list:
            self.executor.sudo(f"ifconfig {port} up")
        delta_time = time.time() + 30
        while time.time() < delta_time:
            try:
                resp = request.urlopen("https://apple.com")
                if resp.code != 404:
                    return
            except error.URLError:
                time.sleep(0.5)

    def turn_internet_off(self):
        for port in self.ports_list:
            self.executor.sudo(f"ifconfig {port} down")

    @property
    def ip_info(self) -> dict:
        """Get external ip information from: ipinfo.io
        :return dict: {'ip': ..., 'hostname': ..., 'city': ..., 'region': ..., 'country': ...}
        """
        for _ in range(5):
            try:
                return json.load(request.urlopen("https://ipinfo.io/json"))
            except error.URLError:
                continue
        return dict()



"""Implements a basic color meter. It finds the value of a colour on given coordinates."""
import multiprocessing
from collections import Counter
from typing import Optional
from typing import Tuple

import cv2

from macuitest.config.colors import CSS3_COLORS
from macuitest.config.constants import Point
from macuitest.lib.elements.ui.monitor import monitor


def get_most_common_color(
    x1: int, y1: int, x2: int, y2: int, ignore_colors: Optional[Tuple[str, ...]] = None
) -> str:
    """Get a color name that is the most common to a specific screen area."""
    pixels = cv2.cvtColor(monitor.make_snapshot(), cv2.COLOR_BGR2RGB)
    with multiprocessing.Pool() as pool:
        colors = pool.map(
            get_closest_color, list(pixels[y, x] for x in range(x1, x2) for y in range(y1, y2))
        )
    if ignore_colors:
        colors = (c for c in colors if c not in ignore_colors)
    return Counter(colors).most_common()[0][0]


def get_color(point: Point) -> str:
    """Get a color name of the given pixel."""
    return get_closest_color(
        cv2.cvtColor(monitor.make_snapshot(), cv2.COLOR_BGR2RGB)[point.y, point.x]
    )


def get_closest_color(pixel) -> str:
    """Calculate RGB difference between `pixel` and every CSS3_COLOR
    and return name of the color with the minimal difference."""
    min_colours = dict()
    pr, pg, pb = pixel
    for name, cr, cg, cb in CSS3_COLORS:
        min_colours[abs(cr - pr) + abs(cg - pg) + abs(cb - pb)] = name
    return min_colours[min(min_colours)]



import time
from datetime import datetime
from datetime import timedelta
from typing import ClassVar

from macuitest.lib.operating_system.shell_executor import ShellExecutor


class TimeManager:
    time_format: ClassVar[str] = "%Y.%m.%d-%H:%M:%S"

    def __init__(self, executor: ShellExecutor, ntp_server: str = "ua.pool.ntp.org"):
        self.ntp_server = ntp_server
        self.executor = executor

    def move_time(self, days: int = 0, hours: int = 0, minutes: int = 5, seconds: int = 0) -> None:
        self.disable_using_network_time()
        start_point: datetime = datetime.now()
        delta = (
            start_point + timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds + 1)
        ).strftime(self.time_format)
        self.executor.sudo(f"date -f {self.time_format} {delta}")
        time.sleep(3)
        elapsed, target_time_shift = (
            datetime.now() - start_point
        ).seconds, hours * 3600 + minutes * 60 + seconds
        if elapsed <= target_time_shift:
            raise EnvironmentError(f"Elapsed: {elapsed}, target: {target_time_shift}")

    def reset_time(self, ntp_server: str = "time.apple.com"):
        """Enable using network time and get it
        by Simple Network Time Protocol from an NTP server."""
        if self.executor.sudo(f"sntp -sS {ntp_server}").returncode != 0:
            self.executor.sudo(f"sntp -s {ntp_server}")
        self.enable_using_network_time()

    @property
    def is_network_time_used(self) -> bool:
        result = (
            self.executor.sudo_get_output("systemsetup -getusingnetworktime")
            .split(":")[-1]
            .strip()
            .lower()
        )
        return {"on": True, "off": False}.get(result)

    def enable_using_network_time(self):
        self.executor.sudo("systemsetup -setusingnetworktime on")

    def disable_using_network_time(self):
        self.executor.sudo("systemsetup -setusingnetworktime off")



import time

import AVFoundation
import Foundation
from Quartz import CGMainDisplayID


class ScreenRecorder:
    def __init__(self):
        self.session = AVFoundation.AVCaptureSession.new()
        self.session.setSessionPreset_(AVFoundation.AVCaptureSessionPresetHigh)
        self.screen_input = AVFoundation.AVCaptureScreenInput.new().initWithDisplayID_(
            CGMainDisplayID()
        )
        self.screen_input.setCapturesMouseClicks_(True)
        self.screen_input.setCapturesCursor_(True)

        if self.session.canAddInput_:
            self.session.addInput_(self.screen_input)

        self.movie_file_output = AVFoundation.AVCaptureMovieFileOutput.new()
        if self.session.canAddOutput_:
            self.session.addOutput_(self.movie_file_output)

    def start_recording(self, destination: str):
        self.session.startRunning()
        destination = Foundation.NSURL.fileURLWithPath_(destination)
        self.movie_file_output.startRecordingToOutputFileURL_recordingDelegate_(destination, self)
        time.sleep(1.5)

    def stop_recording(self):
        time.sleep(0.5)
        self.movie_file_output.stopRecording()
        time.sleep(2)  # For macOS to catch up.
